/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file matrix2.h
 * @brief The header file for 2x2 matrices.
 * @author Kiminori MATSUZAKI
 */

#ifndef _MATRIX2_H_
#define _MATRIX2_H_

#include <iostream>

/**
 * The class for 2 x 2 Matrices. 
 */
template< typename A >
class Matrix2
{
private:
  A data[2][2];
  
public:
  /**
   * a11 a12 \n
   * a21 a22
   */
  Matrix2( ) { };
  Matrix2( A a11, A a12, A a21, A a22 );
  Matrix2( bool b );
  
public:
  A& at( int x, int y );
  const A& at( int x, int y ) const;

  Matrix2< A > operator+( const Matrix2< A >& other ) const;
  Matrix2< A > operator-( const Matrix2< A >& other ) const;
  Matrix2< A > operator*( const Matrix2< A >& other ) const;
  Matrix2< A > operator*( A val ) const;

  bool operator==( const Matrix2< A >& other ) const;

  static const MPI_Datatype get_mpi_type( );
};

template< typename A >
inline std::ostream& operator<<( std::ostream& s, const Matrix2< A >& x );

template< typename A >
inline std::istream& operator>>( std::istream& s, Matrix2< A >& x );

template<>
template< typename A >
class template_mpitype< Matrix2< A > > {
public:
  static const MPI_Datatype type() { return Matrix2< A >::get_mpi_type( ); }
};

#include "matrix2.tpp"

#endif // _MATRIX2_H_
