/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file map_zip_test.cpp
 * @brief Tests of the map and zip skeletons.
 * @author Kento EMOTO
 */
 

#include <iostream>
#include <cmath>
#include <cstdlib>
#include "matrix_skeletons.h"
#include "primitive_functions.h"
using namespace primitive_functions;

class Mulp : public skeleton::unary_function<std::pair<int, int> , int>{
public:
	void operator()(const std::pair<int, int> p, int *r) const {
		*r = p.first * p.second;
	}
};
class Subp : public skeleton::unary_function<std::pair<int, int>, int > {
public:
	void operator()(const std::pair<int, int> p, int *r) const {
		*r = p.first - p.second;
	}
};



int SketoMain( int argc, char **argv )
{
	
	int n = 13;
	int m = 8;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	if(argc > 2){
		m = atoi(argv[2]);
	}


	int myrank = skeleton::rank;
	
	dist_matrix<int> A(n, m, GenAdd<int>());
	dist_matrix<int> B(n, m, GenSub<int>());
	if(myrank==0){
		std::cout << "A = \n";
	}
	A.print(std::cout);
	if(myrank==0){
		std::cout << "B = \n";
	}
	B.print(std::cout);


	dist_matrix< std::pair <int ,int> > *C = matrix_skeletons::zip(&A, &B);

// 	if(myrank==0){
// 		std::cout << "C = zipw A B\n";
// 	}
// 	C->print(std::cout);

	dist_matrix< int > *D = matrix_skeletons::map(Mulp(), C);

	if(myrank==0){
		std::cout << "D = map mul C\n";
	}
	D->print(std::cout);

	
	dist_matrix< int > *E = matrix_skeletons::map(Subp(), C);

	if(myrank==0){
		std::cout << "E = map sub C\n";
	}
	E->print(std::cout);

	delete C;
	delete D;
	delete E;
	
	return 0;
}
