/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file reduce_scan_test.cpp
 * @brief Tests of the reduce and scan skeletons.
 * @author Kento EMOTO
 */


#include <iostream>
#include <cmath>
#include <cstdlib>
#include "matrix_skeletons.h"
#include "primitive_functions.h"
using namespace primitive_functions;


class Addp : public skeleton::unary_function< std::pair<int, int> , int>{
public:
	void operator()( const std::pair<int, int> p, int *r) const {
		*r = p.first + p.second;
	}
};

int SketoMain( int argc, char **argv )
{
	
	int n = 13;
	int m = 8;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	if(argc > 2){
		m = atoi(argv[2]);
	}


	int myrank = skeleton::rank;
	
	dist_matrix<int> A(n, m, GenAdd<int>());
	dist_matrix<int> B(n, m, GenSub<int>());
	if(myrank==0){
		std::cout << "A = \n";
	}
	A.print(std::cout);
	if(myrank==0){
		std::cout << "B = \n";
	}
	B.print(std::cout);


	int *suma = matrix_skeletons::reduce(Add<int>(), Add<int>(), &A);
	int *sumb = matrix_skeletons::reduce(Add<int>(), Add<int>(), &B);
	dist_matrix< std::pair< int, int > > *C = matrix_skeletons::zip(&A, &B);
	dist_matrix< int > *D = matrix_skeletons::map(Addp(), C);
	dist_matrix< int > *E = matrix_skeletons::scan(Add<int>(), Add<int>(), &A);
	dist_matrix< int > *F = matrix_skeletons::scan(Add<int>(), Add<int>(), &B);
	dist_matrix< int > *G = matrix_skeletons::scanr(Add<int>(), Add<int>(), &A);
	dist_matrix< int > *H = matrix_skeletons::scanr(Add<int>(), Add<int>(), &B);
	int *sum = matrix_skeletons::reduce(Add<int>(), Add<int>(), D); 

	
	
	if(myrank==0){
		std::cout << "suma = reduce (+,+) A = " << *suma << "\n";
		std::cout << "sumb = reduce (+,+) B = " << *sumb << "\n";
		std::cout << "sum = reduce (+,+) (map add (zip A B)) = " << *sum << "\n";
		std::cout << "sum - (suma + sumb) = " << *sum - (*suma + *sumb) << "\n";
	}
	if(myrank==0){
		std::cout << "E = scan (+,+) A\n";
	}
	E->print(std::cout);
	if(myrank==0){
		std::cout << "F = scan (+,+) B\n";
	}
	F->print(std::cout);
	if(myrank==0){
		std::cout << "G = scanr (+,+) A\n";
	}
	G->print(std::cout);
	if(myrank==0){
		std::cout << "H = scanr (+,+) B\n";
	}
	H->print(std::cout);
	delete C;
	delete D;
	delete E;
	delete F;
	delete G;
	delete H;
	return 0;
}
