/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file optimizer.mc
 * @brief Optimizer for SkeTo.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#include <iostream>
#include <vector>
#include <string>
#include <utility>
#include "mop.h"
#include "opt_util.mh"
using namespace std;

// for debug
#define debug(x)
const char *partition = "-----------------------------------------------\n";

// global
vector<pair<Ptree**, int> > exp_vector;

// 
class optimizer : public Class {
public:
  void TranslateMemberFunction(Environment*, Member&);
};

Ptree* prep(Ptree*);
Ptree* translate_into_intermediate(Ptree*);
Ptree* optimize(Ptree*, int);
Ptree* translate_from_intermediate(Ptree*);
bool match2(Ptree**, char*, Ptree**, char*, Ptree**);
bool match3(Ptree**, char*, Ptree**, char*, Ptree**, char*, Ptree**);
bool is_dist_list_dec(Ptree*, Ptree*);
int check_scope(Ptree*);
bool check_mid(Ptree*, Ptree*);
bool check_rear(Ptree*, Ptree*, Ptree**);
Ptree* compose_functions(Ptree*, Ptree*, Ptree*);
Ptree* compose_binary_functions(Ptree*, Ptree*, Ptree*);
Ptree* compose_binary_two_functions(Ptree*, Ptree*, Ptree*, Ptree*);
Ptree* produce_functions(Ptree*, Ptree*, Ptree*);
Ptree* resolve_function(Ptree*, Ptree*);

// ᥤ롼
void optimizer::TranslateMemberFunction(Environment *e, Member &m)
{
  Ptree *body = m.FunctionBody();

  // ǥХå
  debug(cout << partition);
  debug(body->Display());

  // 
  Ptree *new_body = prep(body);

  // ɽѴ
  new_body = translate_into_intermediate(new_body);

  // Ŭ
  new_body = optimize(new_body, 0);

  // C++ ˴Ը
  new_body = translate_from_intermediate(new_body);

  // ǥХå
  debug(cout << partition);
  debug(new_body->Display());

  m.SetFunctionBody(new_body);
}

Ptree* prep(Ptree *exp_list)
{
  int length = Ptree::Length(exp_list);
  
  for(int i = 0; i < length; i++){
    Ptree *exp = Ptree::Nth(exp_list, i);
    Ptree *new_exp = exp;
    Ptree *status, *dec_type, *dec_list;
    Ptree *for1, *for2, *for3, *block;

    // for ʸѿ̵뤹
    if(Ptree::Match(exp, "[for ( %? %? ; %? ) %?]", &for1, &for2, &for3, &block)){
      Ptree *new_block = prep(block);
      new_exp = OptUtil::ReadPtree("[for ( %p %p ; %p ) %p]", for1, for2, for3, new_block);
      exp_list = Ptree::ReplaceAll(exp_list, exp, new_exp);
    }

    // 1 ʸʣѿʸʬ
    else if(Ptree::Match(exp, "[%? %? %? ;]", &status, &dec_type, &dec_list)){
      int dec_length = Ptree::Length(dec_list);

      if(dec_length == 1){
	continue;
      }

      Ptree *new_dec = NIL;
      for(int j = 0; j < dec_length; j++){
	Ptree *name = Ptree::Nth(dec_list, j);
	
	// ΥꥹȤ
	if(!Ptree::Eq(name, ",")){
	  new_dec = Ptree::Append(new_dec, OptUtil::ReadPtree("[[%p %p [%p] ;]]", status, dec_type, name));
	}
      }

      exp_list = OptUtil::Insert(exp_list, new_dec, i);
      int next = Ptree::Length(new_dec) - 1;
      length += next;
      i += next;
    }
    else{
      new_exp = prep(exp);
      exp_list = Ptree::ReplaceAll(exp_list, exp, new_exp);
    }
  }
  
  return exp_list;
}

Ptree* translate_into_intermediate(Ptree *exp_list)
{
  int length = Ptree::Length(exp_list);

  for(int i = 0; i < length; i++){
    Ptree *in, *out, *fun, *oplus, *e;
    Ptree *in1, *in2, *for1, *for2, *for3, *block;
    Ptree *exp = Ptree::Nth(exp_list, i);
    Ptree *exp0 = exp, *new_exp = exp;
    Ptree *status, *dec_type, *dec_list;
    bool is_dec = false;

    // ʸν
    if(Ptree::Match(exp, "[%? %? %? ;]", &status, &dec_type, &dec_list)){
      Ptree *node = Ptree::First(dec_list);
      Ptree *name = NIL, *init = node;
      int dec_length = Ptree::Length(node);

      // ѿ̾ꥹȤˤȷ礹
      for(int j = 0; j < dec_length; j++){
	Ptree *x = Ptree::Nth(node, j);

	if(Ptree::Eq(x, "=")){
	  break;
	}
	else{
	  name = Ptree::Append(name, x);
	  init = Ptree::Rest(init);
	}
      }
      exp = OptUtil::ReadPtree("[%p ;]", Ptree::Cons(name, init));
      is_dec = true;
    }
    else{
      status = NIL;
      dec_type = NIL;
    }

    // map
    if(Ptree::Match(exp, "[[%? = [[list_skeletons :: map]([%?, %?])]] ;]", &out, &fun, &in)){
      new_exp = OptUtil::ReadPtree("[%p %p [1 val %p %p [[cin sin nin] [%p] [aout bout gout]]]]", status, dec_type, in, out, fun);
    }
    else if(Ptree::Match(exp, "[[[list_skeletons :: map]([%?, %?, %?])] ;]", &fun, &in, &out)){
      new_exp = OptUtil::ReadPtree("[%p %p [2 ref %p %p [[cin sin nin] [%p] [aout bout gout]]]]", status, dec_type, in, out, fun);
    }
    else if(Ptree::Match(exp, "[[[list_skeletons :: map_ow]([%?, %?])] ;]", &fun, &in)){
      new_exp = OptUtil::ReadPtree("[%p %p [3 val %p %p [[cin sin nin] [%p] [aout bout gout]]]]", status, dec_type, in, in, fun);
    }

    // reduce
    else if(Ptree::Match(exp, "[[%? = [[list_skeletons :: reduce]([%?, %?, %?])]] ;]", &out, &oplus, &e, &in)){
      new_exp = OptUtil::ReadPtree("[%p %p [1 val %p %p [[%p id %p] [id] [aout bout gout]]]]", status, dec_type, in, out, oplus, e);
    }
    else if(Ptree::Match(exp, "[[[list_skeletons :: reduce]([%?, %?, %?, %?])] ;]", &oplus, &e, &in, &out)){
      new_exp = OptUtil::ReadPtree("[%p %p [2 ref %p %p [[%p id %p] [id] [aout bout gout]]]]", status, dec_type, in, out, oplus, e);
    }

    /*
    // prescan
    else if(Ptree::Match(exp, "[[%? = [[list_skeletons :: prescan]([%?, %?, %?])]] ;]", &out, &oplus, &e, &in)){
      new_exp = OptUtil::ReadPtree("[%p %p [1 val %p %p [[cin sin nin] [id] [aps bps gps]]]]", status, dec_type, in, out, oplus, e);
    }
    else if(Ptree::Match(exp, "[[[list_skeletons :: prescan]([%?, %?, %?, %?])] ;]", &oplus, &e, &in, &out)){
      new_exp = OptUtil::ReadPtree("[%p %p [2 ref %p %p [[cin sin nin] [id] [aps bps gps]]]]", status, dec_type, in, out, oplus, e);
    }
    */

    // zip
    else if(Ptree::Match(exp, "[[%? = [[list_skeletons :: zip]([%?, %?])]] ;]", &out, &in1, &in2)){
      new_exp = OptUtil::ReadPtree("[%p %p [1 val [%p %p] %p [[cin sin nin] [id] [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]", status, dec_type, in1, in2, out);
    }
    else if(Ptree::Match(exp, "[[[list_skeletons :: zip]([%?, %?, %?])] ;]", &in1, &in2, &out)){
      new_exp = OptUtil::ReadPtree("[%p %p [2 ref [%p %p] %p [[cin sin nin] [id] [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]", status, dec_type, in1, in2, out);
    }

    // zipwith
    else if(Ptree::Match(exp, "[[%? = [[list_skeletons :: zipwith]([%?, %?, %?])]] ;]", &out, &fun, &in1, &in2)){
      new_exp = OptUtil::ReadPtree("[%p %p [1 val [%p %p] %p [[cin sin nin] [%p] [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]", status, dec_type, in1, in2, out, fun);
    }
    else if(Ptree::Match(exp, "[[[list_skeletons :: zipwith]([%?, %?, %?, %?])] ;]", &fun, &in1, &in2, &out)){
      new_exp = OptUtil::ReadPtree("[%p %p [2 ref [%p %p] %p [[cin sin nin] [%p] [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]", status, dec_type, in1, in2, out, fun);
    }

    // ֥åѴ
    else if(!is_dec){
      new_exp = translate_into_intermediate(exp);
    }

    // ɽִ
    exp_list = Ptree::ReplaceAll(exp_list, exp0, new_exp);
  }

  return exp_list;
}

Ptree* optimize(Ptree *exp_list, int depth)
{
  int length = Ptree::Length(exp_list);

  // ŬǤʤ
  if(length < 1){
    return exp_list;
  }

  // exp_list õΰ֤ФƤ
  exp_vector.push_back(make_pair(&exp_list, 0));

  for( ; exp_vector.back().second < length; exp_vector.back().second++){
    Ptree *exp, *exp0, *exp1, *exp2, *new_exp, *new_fun;
    Ptree *knd0, *knd1, *knd2, *type0, *type1, *type2, *in0, *in1, *in2_0, *in2_1, *out0, *out1, *out2;
    Ptree *cata, *ana, *ana1, *ana2, *psi, *phi, *eta, *eta1, *eta2;
    Ptree *status, *dec_type;

    exp = Ptree::Nth(exp_list, exp_vector.back().second);

    // ƵƤӽФǥ֥åĴ٤
    new_exp = optimize(exp, depth+1);
    exp_list = Ptree::ReplaceAll(exp_list, exp, new_exp);
    
    // cataJ-hyloJ (tau ̵)
    if(match2(&exp_list,
	      "[%* %* [%? %? %? %? [[cin sin nin] %* %*]]]", &exp0,
	      "[%* %* [%? %? %? %? [%* %* [aout bout gout]]]]", &exp1)){
      Ptree::Match(exp0, "[%* %* [%? %? %? %? [[cin sin nin] %? %?]]]", &knd0, &type0, &in0, &out0, &eta2, &psi);
      Ptree::Match(exp1, "[%? %? [%? %? %? %? [%? %? [aout bout gout]]]]", &status, &dec_type, &knd1, &type1, &in1, &out1, &cata, &eta1);

      bool is_zip = false, is_zipwith = false;
      // exp0 = zip or zipwith
      if(Ptree::Equal(psi, OptUtil::ReadPtree("[sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]"))){
	// exp0 = zip
	if(Ptree::Eq(Ptree::First(eta2), ("id"))){
	  is_zip = true;
	}
	// exp0 = zipwith
	else{
	  is_zipwith = true;
	}
      }

      // exp1 = map
      if(Ptree::Equal(cata, OptUtil::ReadPtree("[cin sin nin]"))){
	if(is_zip){
	  new_fun = resolve_function(eta1, type0);
	}
	else if(is_zipwith){
	  new_fun = compose_binary_functions(eta1, eta2, type0);
	}
	else {
	  new_fun = compose_functions(eta1, eta2, type0);
	}
	new_exp = OptUtil::ReadPtree("[%p %p [%p %p %p %p [%p %p %p]]]", status, dec_type, knd1, type0, in0, out1, cata, new_fun, psi);
      }
      // exp1 = reduce
      else{
	if(is_zip){
	  Ptree *oplus, *e;
	  Ptree::Match(cata, "[%? id %?]", &oplus, &e);
	  if(Ptree::Eq(type0, "val")){
	    new_exp = OptUtil::ReadPtree("[%p %p %p = list_skeletons::zip_reduce( %p, %p, %p, %p );]",
					 status, dec_type, out1, oplus, e, Ptree::First(in0), Ptree::Second(in0)); 
	  }
	  else{
	    new_exp = OptUtil::ReadPtree("[list_skeletons::zip_reduce( %p, %p, %p, %p, %p );]",
					 oplus, e, Ptree::First(in0), Ptree::Second(in0), out1);
	  }
	}
	else if(is_zipwith){
	  Ptree *oplus, *e;
	  Ptree::Match(cata, "[%? id %?]", &oplus, &e);
	  if(Ptree::Eq(type0, "val")){
	    new_exp = OptUtil::ReadPtree("[%p %p %p = list_skeletons::zipwith_reduce( %p, %p, %p, %p, %p );]",
					 status, dec_type, out1, eta2, oplus, e, Ptree::First(in0), Ptree::Second(in0)); 
	  }
	  else{
	    new_exp = OptUtil::ReadPtree("[list_skeletons::zipwith_reduce( %p, %p, %p, %p, %p, %p );]",
					 eta2, oplus, e, Ptree::First(in0), Ptree::Second(in0), out1); 
	  }
	}
	else{
	  new_exp = OptUtil::ReadPtree("[%p %p [%p %p %p %p [%p %p %p]]]", status, dec_type, knd1, type0, in0, out1, cata, eta2, psi);
	}
      }

      exp_list = Ptree::ReplaceAll(exp_list, exp1, new_exp);
    }

    // hyloj-anaj*anaj (phi ̵롤ana  out (ǽ2Ĥ zipzipwith ʤ))
    else if(match3(&exp_list,
	           "[%* %* [%? %? %? %? [[cin sin nin] %* [aout bout gout]]]]", &exp0,
		   "[%* %* [%? %? %? %? [[cin sin nin] %* [aout bout gout]]]]", &exp1,
		   "[%* %* [%? %? [%? %?] %? [[cin sin nin] %* [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]", &exp2)){
      Ptree::Match(exp0, "[%* %* [%? %? %? %? [[cin sin nin] %? [aout bout gout]]]]", &knd0, &type0, &in0, &out0, &eta1);
      Ptree::Match(exp1, "[%* %* [%? %? %? %? [[cin sin nin] %? [aout bout gout]]]]", &knd1, &type1, &in1, &out1, &eta2);
      Ptree::Match(exp2, "[%? %? [%? %? [%? %?] %? [[cin sin nin] %? [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]",
		   &status, &dec_type, &knd2, &type2, &in2_0, &in2_1, &out2, &eta);
      // zip
      if(Ptree::Eq(Ptree::First(eta), "id")){
	new_fun = produce_functions(eta1, eta2, type0);
      }
      // zipwith
      else{
	new_fun = compose_binary_two_functions(eta, eta1, eta2, type0);
      }
      new_exp = OptUtil::ReadPtree("[%p %p [%p %p [%p %p] %p [[cin sin nin] %p [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]", status, dec_type, knd2, type2, in0, in1, out2, new_fun);
      exp_list = Ptree::ReplaceAll(exp_list, exp2, new_exp);
    }
  }
  
  // exp_list õΰ֤ΤƤ
  exp_vector.pop_back();

  return exp_list;
}

Ptree* translate_from_intermediate(Ptree *exp_list)
{
  int length = Ptree::Length(exp_list);
  
  for(int i = 0; i < length; i++){
    Ptree *kind, *in, *in1, *in2, *out;
    Ptree *fun, *oplus, *p, *e;
    Ptree *exp = Ptree::Nth(exp_list, i);
    Ptree *new_exp = exp;
    Ptree *status, *dec_type;

    // map
    if(Ptree::Match(exp, "[%? %? [%? %* %? %? [[cin sin nin] %? [aout bout gout]]]]", &status, &dec_type, &kind, &in, &out, &fun)){
      if(Ptree::Eq(kind, "1")){
	new_exp = Ptree::qMake("`status` `dec_type` `out` = list_skeletons::map(`fun`, `in`);");
      }
      else if(Ptree::Eq(kind, "2")){
	new_exp = Ptree::qMake("`status` `dec_type` list_skeletons::map(`fun`, `in`, `out`);");
      }
      else if(Ptree::Eq(kind, "3")){
	new_exp = Ptree::qMake("`status` `dec_type` list_skeletons::map_ow(`fun`, `in`);");
      }
    }

    // reduce
    else if(Ptree::Match(exp, "[%? %? [%? %* %? %? [[%? id %?] [id] [aout bout gout]]]]", &status, &dec_type, &kind, &in, &out, &oplus, &e)){
      if(Ptree::Eq(kind, "1")){
	new_exp = Ptree::qMake("`status` `dec_type` `out` = list_skeletons::reduce(`oplus`, `e`, `in`);");
      }
      else if(Ptree::Eq(kind, "2")){
	new_exp = Ptree::qMake("`status` `dec_type` list_skeletons::reduce(`oplus`, `e`, `in`, `out`);");
      }
    }

    /*
    // prescan
    else if(Ptree::Match(exp, "[%? %? [%? %* %? %? [[cin sin nin] [id] [aps bps gps]]]]", &status, &dec_type, &kind, &in, &out, &oplus, &e)){
      if(Ptree::Eq(kind, "1")){
	new_exp = Ptree::qMake("`status` `dec_type` `out` = list_skeletons::prescan(`oplus`, `e`, `in`);");
      }
      else if(Ptree::Eq(kind, "2")){
	new_exp = Ptree::qMake("`status` `dec_type` list_skeletons::prescan(`oplus`, `e`, `in`, `out`);");
      }
    }
    */

    // zip
    else if(Ptree::Match(exp, "[%? %? [%? %* [%? %?] %? [[cin sin nin] [id] [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]", &status, &dec_type, &kind, &in1, &in2, &out)){
      if(Ptree::Eq(kind, "1")){
	new_exp = Ptree::qMake("`status` `dec_type` `out` = list_skeletons::zip(`in1`, `in2`);");
      }
      else if(Ptree::Eq(kind, "2")){
	new_exp = Ptree::qMake("`status` `dec_type` list_skeletons::zip(`in1`, `in2`, `out`);");
      }
    }

    // zipwith
    else if(Ptree::Match(exp, "[%? %? [%? %* [%? %?] %? [[cin sin nin] %? [sigma [aout1 bout1 gout1] [aout2 bout2 gout2]]]]]", &status, &dec_type, &kind, &in1, &in2, &out, &fun)){
      if(Ptree::Eq(kind, "1")){
	new_exp = Ptree::qMake("`status` `dec_type` `out` = list_skeletons::zipwith(`fun`, `in1`, `in2`);");
      }
      else if(Ptree::Eq(kind, "2")){
	new_exp = Ptree::qMake("`status` `dec_type` list_skeletons::zipwith(`fun`, `in1`, `in2`, `out`);");
      }
    }

    // cataj
    else if(Ptree::Match(exp, "[%? %? [%? %* %? %? [[%? id %?] %? [aout bout gout]]]]", &status, &dec_type, &kind, &in, &out, &oplus, &e, &fun)){
      if(Ptree::Eq(kind, "1")){
	new_exp = Ptree::qMake("`status` `dec_type` `out` = list_skeletons::cataj(`oplus`, `fun`, `e`, `in`);");
      }
      else if(Ptree::Eq(kind, "2")){
	new_exp = Ptree::qMake("`status` `dec_type` list_skeletons::cataj(`oplus`, `fun`, `e`, `in`, `out`);");
      }
    }
    
    // ֥åԸ
    else{
      new_exp = translate_from_intermediate(exp);
    }

    // C++ Υɤִ
    exp_list = Ptree::ReplaceAll(exp_list, exp, new_exp);
  }

  return exp_list;
}

bool match2(Ptree **exp_list, char *fst, Ptree **m_exp0, char *snd, Ptree **m_exp1)
{
  bool two_input = false;
  int length = Ptree::Length(*exp_list);
  int index_bak = exp_vector.back().second;
  Ptree *exp0 = Ptree::Nth(*exp_list, index_bak), *exp1;
  Ptree *knd0, *knd1, *type0, *type1, *in0, *in1, *out0, *out1;

  // 1ܤʸ˥ޥåʤ
  if(!Ptree::Match(exp0, fst, &knd0, &type0, &in0, &out0)){
    return false;
  }

  // ϥꥹȤ2ĤȽǤ"*", "&" 
  Ptree *in0_name0, *in0_name1;
  Ptree *out0_name = OptUtil::DeleteUnary(out0);
  if((Ptree::Length(in0) == 2) &&
     !Ptree::Eq(Ptree::First(in0), "*") && !Ptree::Eq(Ptree::First(in0), "&")){
    two_input = true;
    in0_name0 = OptUtil::DeleteUnary(Ptree::First(in0));
    in0_name1 = OptUtil::DeleteUnary(Ptree::Second(in0));
    // in ХѿǤʤĴ٤
    if(check_scope(in0_name0) == -1 || check_scope(in0_name1) == -1){
      return false;
    }
  }
  else{
    in0_name0  = OptUtil::DeleteUnary(in0);
    // in ХѿǤʤĴ٤
    if(check_scope(in0_name0) == -1){
      return false;
    }
  }

  // 2ܤʸ˥ޥå뤫
  for(exp_vector.back().second++; ; exp_vector.back().second++){
    // 2ܤʸĤʤä
    if(exp_vector.back().second >= length){
      exp_vector.back().second = index_bak;
      return false;
    }
    exp1 = Ptree::Nth(*exp_list, exp_vector.back().second);
    // 2ܤʸ˥ޥå
    if(Ptree::Match(exp1, snd, &knd1, &type1, &in1, &out1) &&
       Ptree::Eq(type0, type1) &&
       Ptree::Equal(out0_name, OptUtil::DeleteUnary(in1))){
      break;
    }
    // 1ܤ2ܤδ֤ in0out0 ޤޤƤͻԲǽ
    else if(!(check_mid(Ptree::List(exp1), in0_name0) &&
	      check_mid(Ptree::List(exp1), out0_name) &&
	      (!two_input || check_mid(Ptree::List(exp1), in0_name1)))){
      exp_vector.back().second = index_bak;
      return false;
    }
  }

  // out0  out1 ʤȤϡ2ܰʹߤʸĴ٤
  if(!Ptree::Eq(out0, out1)){
    // out0 ΥפĴ٤
    Ptree *name;
    int scope_index;
    if(Ptree::Match(exp0, "[%* [dist_list %*] [%* %* %* %? %*]]", &name) &&
       OptUtil::IsInclude(name, out0_name)){
      scope_index = exp_vector.size() - 1;
    }
    else{
      scope_index = check_scope(out0_name);
    }

    if(scope_index == -1){
      exp_vector.back().second = index_bak;
      return false;
    }

    // out0  delete ʸ򸡺
    int delete_index = -1;
    Ptree *delete_exp = NIL;
    for(int i = exp_vector.size()-1; i >= scope_index; i--){
      Ptree **v_exp_list = exp_vector[i].first;
      int v_length = Ptree::Length(*v_exp_list);
      for(int j = exp_vector[i].second+1; j < v_length; j++){
	Ptree *v_exp = Ptree::Nth(*v_exp_list, j);
	if(!check_rear(Ptree::List(v_exp), out0_name, &delete_exp)){
	  exp_vector.back().second = index_bak;
	  return false;
	}
	else if((delete_index == -1) && !Ptree::Eq(delete_exp, NIL)){
	  delete_index = i;
	}
      }
    }
    
    // in0 == out0 뤤 kind == ref Ǥʤ delete ʸ
    if(delete_index != -1 && !Ptree::Eq(knd0, "ref") && !Ptree::Eq(in0, out0)){
      Ptree **exp = exp_vector[delete_index].first;
      *exp = Ptree::ReplaceAll(*exp, delete_exp, NIL);
    }
  }
  
  *m_exp0 = exp0;
  *m_exp1 = exp1;
  exp_vector.back().second = index_bak;
  *exp_list = Ptree::ReplaceAll(*exp_list, exp0, NIL); // 1ʸܤʸˤ

  return true;
}

bool match3(Ptree **exp_list, char *fst, Ptree **m_exp0, char *snd, Ptree **m_exp1, char *thd, Ptree **m_exp2)
{
  bool two_input0 = false, two_input1 = false;
  int length = Ptree::Length(*exp_list);
  int index_bak = exp_vector.back().second;
  Ptree *exp0 = Ptree::Nth(*exp_list, index_bak), *exp1, *exp2;
  Ptree *knd0, *knd1, *knd2, *type0, *type1, *type2, *in0, *in1, *in2_0, *in2_1, *out0, *out1, *out2;

  // 1ܤʸ˥ޥåʤ
  if(!Ptree::Match(exp0, fst, &knd0, &type0, &in0, &out0)){
    return false;
  }

  // ϥꥹȤ2ĤȽǤ"*", "&" 
  Ptree *in0_name0, *in0_name1;
  Ptree *out0_name = OptUtil::DeleteUnary(out0);
  if((Ptree::Length(in0) == 2) &&
     !Ptree::Eq(Ptree::First(in0), "*") && !Ptree::Eq(Ptree::First(in0), "&")){
    two_input0 = true;
    in0_name0 = OptUtil::DeleteUnary(Ptree::First(in0));
    in0_name1 = OptUtil::DeleteUnary(Ptree::Second(in0));
    // in ХѿǤʤĴ٤
    if(check_scope(in0_name0) == -1 || check_scope(in0_name1) == -1){
      return false;
    }
  }
  else{
    in0_name0  = OptUtil::DeleteUnary(in0);
    // in ХѿǤʤĴ٤
    if(check_scope(in0_name0) == -1){
      return false;
    }
  }

  // 2ܤʸ˥ޥå뤫
  for(exp_vector.back().second++; ; exp_vector.back().second++){
    // 2ܤʸĤʤä
    if(exp_vector.back().second >= length){
      exp_vector.back().second = index_bak;
      return false;
    }
    exp1 = Ptree::Nth(*exp_list, exp_vector.back().second);
    // 2ܤʸ˥ޥå
    if(Ptree::Match(exp1, snd, &knd1, &type1, &in1, &out1) && Ptree::Eq(type0, type1) &&
       !Ptree::Equal(OptUtil::DeleteUnary(out1), in0_name0) &&
       !(two_input0 && Ptree::Equal(OptUtil::DeleteUnary(out1), in0_name1)) &&
       !Ptree::Equal(OptUtil::DeleteUnary(out1), out0_name) &&
       !Ptree::Equal(OptUtil::DeleteUnary(in1), out0_name)){
      break;
    }
    // 1ܤ2ܤδ֤ in0out0 ޤޤƤͻԲǽ
    else if(!(check_mid(Ptree::List(exp1), in0_name0) &&
	      check_mid(Ptree::List(exp1), out0_name) &&
	      (!two_input0 || check_mid(Ptree::List(exp1), in0_name1)))){
      exp_vector.back().second = index_bak;
      return false;
    }
  }

  // ϥꥹȤ2ĤȽǤ"*", "&" 
  Ptree *in1_name0, *in1_name1;
  Ptree *out1_name = OptUtil::DeleteUnary(out1);
  if((Ptree::Length(in1) == 2) &&
     !Ptree::Eq(Ptree::First(in1), "*") && !Ptree::Eq(Ptree::First(in1), "&")){
    two_input1 = true;
    in1_name0 = OptUtil::DeleteUnary(Ptree::First(in1));
    in1_name1 = OptUtil::DeleteUnary(Ptree::Second(in1));
    // in ХѿǤʤĴ٤
    if(check_scope(in1_name0) == -1 || check_scope(in1_name1) == -1){
      return false;
    }
  }
  else{
    in1_name0  = OptUtil::DeleteUnary(in1);
    // in ХѿǤʤĴ٤
    if(check_scope(in1_name0) == -1){
      return false;
    }
  }

  // 3ܤʸ˥ޥå뤫
  for(exp_vector.back().second++; ; exp_vector.back().second++){
    // 3ܤʸĤʤä
    if(exp_vector.back().second >= length){
      exp_vector.back().second = index_bak;
      return false;
    }
    exp2 = Ptree::Nth(*exp_list, exp_vector.back().second);
    // 3ܤʸ˥ޥå
    if(Ptree::Match(exp2, thd, &knd2, &type2, &in2_0, &in2_1, &out2) &&
       Ptree::Equal(out0_name, OptUtil::DeleteUnary(in2_0)) &&
       Ptree::Equal(out1_name, OptUtil::DeleteUnary(in2_1))){
      break;
    }
    // 2ܤ3ܤδ֤ in0out0in1out1 ޤޤƤͻԲǽ
    else if(!(check_mid(Ptree::List(exp1), in0_name0) &&
	      check_mid(Ptree::List(exp1), out0_name) &&
	      (!two_input0 || check_mid(Ptree::List(exp1), in0_name1)) &&
	      check_mid(Ptree::List(exp1), in1_name0) &&
	      check_mid(Ptree::List(exp1), out1_name) &&
	      (!two_input1 || check_mid(Ptree::List(exp1), in1_name1)))){
      exp_vector.back().second = index_bak;
      return false;
    }
  }

  // out0out1 ΥפĴ٤
  Ptree *name;
  int scope_index0;
  if(Ptree::Match(exp0, "[%* [dist_list %*] [%* %* %* %? %*]]", &name) &&
     OptUtil::IsInclude(name, out0_name)){
    scope_index0 = exp_vector.size() - 1;
  }
  else{
    scope_index0 = check_scope(out0_name);
  }

  int scope_index1;
  if(Ptree::Match(exp1, "[%* [dist_list %*] [%* %* %* %? %*]]", &name) &&
     OptUtil::IsInclude(name, out0_name)){
    scope_index1 = exp_vector.size() - 1;
  }
  else{
    scope_index1 = check_scope(out0_name);
  }

  if(scope_index0 == -1 || scope_index1 == -1){
    exp_vector.back().second = index_bak;
    return false;
  }

  // out0  delete ʸ򸡺
  int delete_index0 = -1;
  Ptree *delete_exp0 = NIL;
  for(int i = exp_vector.size()-1; i >= scope_index0; i--){
    Ptree **v_exp_list = exp_vector[i].first;
    int v_length = Ptree::Length(*v_exp_list);
    for(int j = exp_vector[i].second+1; j < v_length; j++){
      Ptree *v_exp = Ptree::Nth(*v_exp_list, j);
      if(!check_rear(Ptree::List(v_exp), out0_name, &delete_exp0)){
	exp_vector.back().second = index_bak;
	return false;
      }
      else if(delete_index0 == -1 && !Ptree::Eq(delete_exp0, NIL)){
	delete_index0 = i;
      }
    }
  }
  
  // out1  delete ʸ򸡺
  int delete_index1 = -1;
  Ptree *delete_exp1 = NIL;
  for(int i = exp_vector.size()-1; i >= scope_index1; i--){
    Ptree **v_exp_list = exp_vector[i].first;
    int v_length = Ptree::Length(*v_exp_list);
    for(int j = exp_vector[i].second+1; j < v_length; j++){
      Ptree *v_exp = Ptree::Nth(*v_exp_list, j);
      if(!check_rear(Ptree::List(v_exp), out1_name, &delete_exp1)){
	exp_vector.back().second = index_bak;
	return false;
      }
      else if(delete_index1 == -1 && !Ptree::Eq(delete_exp1, NIL)){
	delete_index1 = i;
      }
    }
  }
  
  // in0 == out0 뤤 kind == ref Ǥʤ delete ʸ
  if(delete_index0 != -1 && !Ptree::Eq(knd0, "ref") && !Ptree::Eq(in0, out0)){
    Ptree **exp = exp_vector[delete_index0].first;
    *exp = Ptree::ReplaceAll(*exp, delete_exp0, NIL);
  }
  // in1 == out1 뤤 kind == ref Ǥʤ delete ʸ
  if(delete_index1 != -1 && !Ptree::Eq(knd0, "ref") && !Ptree::Eq(in1, out1)){
    Ptree **exp = exp_vector[delete_index1].first;
    *exp = Ptree::ReplaceAll(*exp, delete_exp1, NIL);
  }

  *m_exp0 = exp0;
  *m_exp1 = exp1;
  *m_exp2 = exp2;
  exp_vector.back().second = index_bak;
  *exp_list = Ptree::ReplaceAll(*exp_list, exp0, NIL); // 1ʸܤʸˤ
  *exp_list = Ptree::ReplaceAll(*exp_list, exp1, NIL); // 2ʸܤʸˤ
  return true;
}

bool is_dist_list_dec(Ptree *exp, Ptree *_name)
{
  Ptree *name, *name_list;
  // dist_list ȯ
  if(Ptree::Match(exp, "[%* [dist_list %*] %?;]", &name_list)){
    int n_len = Ptree::Length(name_list);
    // ѿ̾Ĵ٤
    for(int i = 0; i < n_len; i++){
      Ptree *x = Ptree::Nth(name_list, i);
      if(!Ptree::Eq(x, ",")){
	// ݥ󥿤äȤн
	int x_len = Ptree::Length(x);
	for(int j = 0; j < x_len; j++){
	  name = Ptree::Nth(x, j);
	  if(!Ptree::Eq(name, "*")){
	    break;
	  }
	}
	// ѿ̾ȯ
	if(Ptree::Eq(name, _name)){
	  return true;
	}
      }
    }
  }
  else if(Ptree::Match(exp, "[%* [dist_list %*] [%* %* %* %? %*]]", &name) &&
	  OptUtil::IsInclude(name, _name)){
    return true;
  }
  // dist_list ǤϤʤ
  return false;
}

int check_scope(Ptree *name)
{
  // name Ǥξ硤ͻǤʤΤ -1 ֤
  // ߤΰ֤餵ΤܤäĴ٤
  for(int i = exp_vector.size()-1; i >=0; i--){
    Ptree **v_exp = exp_vector[i].first;
    for(int j = exp_vector[i].second-1; j >=0; j--){
      if(is_dist_list_dec(Ptree::Nth(*v_exp, j), name)){
	return i;
      }
    }
  }
  return -1; // Ĥʤä
}

bool check_mid(Ptree *exp_list, Ptree *name)
{
  if(Ptree::Eq(exp_list, NIL)){
    return true;
  }
  else if(exp_list->IsLeaf()){
    return !Ptree::Eq(exp_list, name);
  }

  int length = Ptree::Length(exp_list);

  for(int i = 0; i < length; i++){
    Ptree *exp = Ptree::Nth(exp_list, i);

    // ƵƤӽФ
    if(check_mid(exp, name)){
      continue;
    }
    // Ʊѿ̾СʹߤʸĴ٤ʤƤ褤
    else if(is_dist_list_dec(exp, name)){
      return true;
    }
    // name ȤƤΤͻԲǽ
    else{
      return false;
    }
  }
  // name ϻȤƤʤä
  return true;
}

bool check_rear(Ptree *exp_list, Ptree *name, Ptree **delete_exp)
{
  if(Ptree::Eq(exp_list, NIL)){
    return true;
  }
  else if(exp_list->IsLeaf()){
    return !Ptree::Eq(exp_list, name);
  }

  int length = Ptree::Length(exp_list);

  for(int i = 0; i < length; i++){
    Ptree *x;
    Ptree *exp = Ptree::Nth(exp_list, i);

    // ƵƤӽФ
    if(check_rear(exp, name, delete_exp)){
      continue;
    }
    // Ʊѿ̾СʹߤʸĴ٤ʤƤ褤
    else if(is_dist_list_dec(exp, name)){
      return true;
    }
    // delete name; 
    else if(Ptree::Eq(*delete_exp, NIL) &&
	    Ptree::Match(exp, "[[delete %?];]", &x) &&
	    OptUtil::IsInclude(x, name)){
      *delete_exp = exp;
      return true;
    }
    // name ȤƤΤͻԲǽ
    else{
      return false;
    }
  }

  // name ϻȤƤʤä
  return true;
}

Ptree* compose_functions(Ptree *f, Ptree *g, Ptree *type)
{
  // id ̵뤹
  if(Ptree::Eq(Ptree::First(f), ("id"))){
    return g;
  }
  // id ̵뤹
  else if(Ptree::Eq(Ptree::First(g), ("id"))){
    return f;
  }
  else if(Ptree::Eq(type, "val")){
    return Ptree::qMake("skeleton::compose_unary_functions(`f`, `g`)");
  }
  else{
    return Ptree::qMake("skeleton::compose_unary_ptr_functions(`f`, `g`)");
  }    
}

Ptree* compose_binary_functions(Ptree *f, Ptree *g, Ptree *type)
{
  // id ̵뤹
  if(Ptree::Eq(Ptree::First(f), ("id"))){
    return g;
  }
  else if(Ptree::Eq(type, "val")){
    return Ptree::qMake("skeleton::compose_binary_functions(`f`, `g`)");
  }
  else{
    return Ptree::qMake("skeleton::compose_binary_ptr_functions(`f`, `g`)");
  }
}

Ptree* compose_binary_two_functions(Ptree *f, Ptree *g, Ptree *h, Ptree *type)
{
  if(Ptree::Eq(type, "val")){
    return Ptree::qMake("skeleton::compose_binary_two_functions(`f`, `g`, `h`)");
  }
  else{
    return Ptree::qMake("skeleton::compose_binary_two_ptr_functions(`f`, `g`, `h`)");
  }
}

Ptree* produce_functions(Ptree *f, Ptree *g, Ptree *type)
{
  if(Ptree::Eq(type, "val")){
    return Ptree::qMake("skeleton::produce_functions(`f`, `g`)");
  }
  else{
    return Ptree::qMake("skeleton::produce_ptr_functions(`f`, `g`)");
  }
}

Ptree* resolve_function(Ptree *f, Ptree *type)
{
  if(Ptree::Eq(type, "val")){
    return Ptree::qMake("skeleton::resolve_function(`f`)");
  }
  else{
    return Ptree::qMake("skeleton::resolve_ptr_function(`f`)");
  }
}
