metaclass optimizer Main;
#include "list_skeletons.h"
using namespace std;

const int SIZE = 10;

struct Inc : public skeleton::unary_function<int, int> {
  int operator()(int x) const { return x + 1; }
} inc;

struct IncPtr : public skeleton::unary_ptr_function<int, int> {
  void operator()(int *r, const int *x) const { *r =  *x + 1; }
} incptr;

struct IncPair : public skeleton::unary_function<pair<int, int>, pair<int, int> > {
  pair<int, int> operator()(pair<int, int> x) const { return make_pair(x.first+1, x.second+1); }
} incpair;

struct Plus : public skeleton::binary_function<int, int, int> {
  int operator()(int x, int y) const { return x + y; }
} plus;
const int unit = 0;

struct PlusPair : public skeleton::binary_function<pair<int, int>, pair<int, int>, pair<int, int> > {
  pair<int, int> operator()(pair<int, int> x, pair<int, int> y) const
  { return make_pair(x.first+y.first, x.second+y.second); }
} plus_pair;
const pair<int, int> p_unit = make_pair(0, 0);

struct Gen : public skeleton::unary_function<int, int> {
  int operator()(int i) const { return i; }
} gen;

skeleton::Snd<int> snd;

ostream& operator<<(ostream &os, const std::pair<int, int> &p)
{
  os << "(" << p.first << "," << p.second << ")";
  return os;
}

int Main::main(int argc, char **argv)
{
  /****************
   * CataJ-HyloJ *
   ****************/

  // map map -> map
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = list_skeletons::map(inc, as);
    dist_list<int> *cs = list_skeletons::map(inc, bs);

    delete as;
    delete bs;
    delete cs;
  }

  // map reduce -> cataj
  {
    cout << "----------\n";
    int result;
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = list_skeletons::map(inc, as);
    result = list_skeletons::reduce(plus, unit, bs);

    delete as;
    delete bs;
  }

  // zip map -> zipwith
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = new dist_list<int>(gen, SIZE);
    dist_list<pair<int, int> > *cs = list_skeletons::zip(as, bs);
    dist_list<pair<int, int> > *ds = list_skeletons::map(incpair, cs);

    delete as;
    delete bs;
    delete cs;
    delete ds;
  }

  // zipwith map -> zipwith
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = new dist_list<int>(gen, SIZE);
    dist_list<int> *cs = list_skeletons::zipwith(plus, as, bs);
    dist_list<int> *ds = list_skeletons::map(inc, cs);

    delete as;
    delete bs;
    delete cs;
    delete ds;
  }

  // zip reduce -> zip_reduce
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = new dist_list<int>(gen, SIZE);
    dist_list<pair<int, int> > *cs = list_skeletons::zip(as, bs);
    pair<int, int> result = list_skeletons::reduce(plus_pair, p_unit, cs);

    delete as;
    delete bs;
    delete cs;
  }

  // zipwith reduce -> zipwith_reduce
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = new dist_list<int>(gen, SIZE);
    dist_list<int> *cs = list_skeletons::zipwith(plus, as, bs);
    int result = list_skeletons::reduce(plus, unit, cs);

    delete as;
    delete bs;
    delete cs;
  }

  /*******************
   * HyloJ-AnaJ*AnaJ *
   *******************/

  // map map zip -> zipwith
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = new dist_list<int>(gen, SIZE);
    dist_list<int> *cs = list_skeletons::map(inc, as);
    dist_list<int> *ds = list_skeletons::map(inc, bs);
    dist_list<pair<int, int> > *es = list_skeletons::zip(cs, ds);

    delete as;
    delete bs;
    delete cs;
    delete ds;
    delete es;
  }

  // map map zipwith -> zipwith
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = new dist_list<int>(gen, SIZE);
    dist_list<int> *cs = list_skeletons::map(inc, as);
    dist_list<int> *ds = list_skeletons::map(inc, bs);
    dist_list<int> *es = list_skeletons::zipwith(plus, cs, ds);

    delete as;
    delete bs;
    delete cs;
    delete ds;
    delete es;
  }

  /****************************
   * $B:GE,2=5,B'0J30$N%A%'%C%/(B *
   ****************************/
  // dist_list $B$,%]%$%s%?$G$J$$>l9g(B
  {
    cout << "----------\n";
    dist_list<int> as(gen, SIZE);
    dist_list<int> bs(SIZE);
    dist_list<int> cs(SIZE);
    list_skeletons::map(incptr, &as, &bs);
    list_skeletons::map(incptr, &bs, &cs);
  }

  // $B4V$K0c$&%9%3!<%W(B
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = list_skeletons::map(inc, as);

    for(int i = 0; i < 0; i++){
      dist_list<int> *bs;
    }
    
    dist_list<int> *cs = list_skeletons::map(inc, bs);
    
    delete as;
    delete bs;
    delete cs;
  }

  // $B2<$K0c$&%9%3!<%W(B
  {
    cout << "----------\n";
    dist_list<int> *as = new dist_list<int>(gen, SIZE);
    dist_list<int> *bs = list_skeletons::map(inc, as);
    dist_list<int> *cs = list_skeletons::map(inc, bs);
    
    for(int i = 0; i < 0; i++){
      dist_list<int> *bs;
    }
    
    delete as;
    delete bs;
    delete cs;
  }

  return 0;
}
