/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file dist_rose_tree.h
 * @brief The Header of class @a dist_rose_tree, the data for distributed
 *        general trees and read/write functions. 
 * @author Kiminori MATSUZAKI
 */

#ifndef __DIST_ROSE_TREE_H__
#define __DIST_ROSE_TREE_H__

#include "dist_tree.h"
#include <fstream>

class rose_tree_skeletons;
class rose_tree_skeletons_impl;

/**
 * @brief A template class for handling rose trees
 *        distributed on processors.
 *
 * This parallel data structure is implemented using the
 * parallel data structure for binary trees.
 */
template< typename A >
class dist_rose_tree
{
  friend class rose_tree_skeletons;
  friend class rose_tree_skeletons_impl;
  
private:
  /** the instance in the binary tree representation. */
  dist_tree< A >* btree;

  /**
   * @brief Default Constructor.
   *
   * To restrict the construction, the default constructor is
   * defined as a private function.
   */
private:
  dist_rose_tree( );

  /**
   * @brief Constructor from the binary tree representation.
   *
   * This constructor generates a rose tree from the binary tree
   * representation. 
   */
private:
  dist_rose_tree( dist_tree< A >* btree_ );

  /**
   * @brief Destructor
   */
public:
  ~dist_rose_tree( );

  /**
   * @todo write comments
   */
public:
  static dist_rose_tree< A >*
  read_from_file( const char* filename );

  /**
   * @todo write comments
   */
public:
  bool write_to_file( const char *filename ) const;

  /**
   * @brief Simple Print function.
   *
   * This function is only for debugging, because the order of output
   * is not guaranteed. 
   */
public:
  void print( ) const;

public:
  void setroot( const A& val );
};

#include "dist_rose_tree.tpp"

#endif // __DIST_TREE_H__
