/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file rose_tree_skeletons.h
 * @brief The header file of class @a rose_tree_skeletons;
 *        a collection of skeletons on rose trees. 
 * @author Kiminori MATSUZAKI
 */

#ifndef __ROSE_TREE_SKELETONS_H__
#define __ROSE_TREE_SKELETONS_H__

#include "dist_rose_tree.h"
#include "tree_skeletons.h"

/**
 * @brief A collection of parallel skeletons for rose trees.
 *
 */
class rose_tree_skeletons
{
  /**
   * @todo write comments
   */
public:
  template< typename K >
  static dist_rose_tree< typename K::result_type > *
  map( const K &k,
       const dist_rose_tree< typename K::argument_type > *t );

  /**
   * @todo write comments
   */
public:
  template< typename K >
  static dist_rose_tree< typename K::result_type > *
  zipwith( const K &k,
	   const dist_rose_tree< typename K::first_argument_type > *t1,
	   const dist_rose_tree< typename K::second_argument_type > *t2);

  /**
   * @todo write comments
   */
public:
  template< typename OPLUS, typename OTIMES >
  static typename OTIMES::result_type
  reduce_ring( const OPLUS &oplus,
	       const typename OPLUS::first_argument_type &unit_oplus,
	       const OTIMES &otimes,
	       const typename OTIMES::first_argument_type &unit_otimes,
	       const dist_rose_tree< typename OPLUS::first_argument_type > *t );

  /**
   * @todo write comments
   */
public:
  template< typename OPLUS, typename OTIMES,
	    typename P1, typename P2, typename P3 >
  static typename OPLUS::result_type
  reduce_exdist( const OPLUS &oplus,
		 const typename OPLUS::first_argument_type &unit_oplus,
		 const OTIMES &otimes,
		 const typename OTIMES::first_argument_type &unit_otimes,
		 const P1 &p1,
		 const P2 &p2,
		 const P3 &p3,
		 const dist_rose_tree< typename OPLUS::first_argument_type > *t );

  /**
   * @todo write comments.
   */
public:
  template< typename OPLUS >
  static typename OPLUS::result_type
  reduce_assoc( const OPLUS &oplus,
		const typename OPLUS::first_argument_type &unit_oplus,
		const dist_rose_tree< typename OPLUS::first_argument_type > *t );
  
  /**
   * @todo write comments
   */
public:
  template< typename OPLUS, typename OTIMES >
  static dist_rose_tree< typename OTIMES::result_type > *
  uAcc_ring( const OPLUS &oplus,
	     const typename OPLUS::first_argument_type &unit_oplus,
	     const OTIMES &otimes,
	     const typename OTIMES::first_argument_type &unit_otimes,
	     const dist_rose_tree< typename OPLUS::first_argument_type > *t );

  /**
   * @todo write comments
   */
public:
  template< typename OPLUS, typename OTIMES,
	    typename P1, typename P2, typename P3 >
  static dist_rose_tree< typename OPLUS::result_type > *
  uAcc_exdist( const OPLUS &oplus,
		 const typename OPLUS::first_argument_type &unit_oplus,
		 const OTIMES &otimes,
		 const typename OTIMES::first_argument_type &unit_otimes,
		 const P1 &p1,
		 const P2 &p2,
		 const P3 &p3,
		 const dist_rose_tree< typename OPLUS::first_argument_type > *t );

  /**
   * @todo write comments
   */
public:
  template< typename OPLUS >
  static dist_rose_tree< typename OPLUS::result_type > *
  uAcc_assoc( const OPLUS &oplus,
	     const typename OPLUS::first_argument_type &unit_oplus,
	     const dist_rose_tree< typename OPLUS::first_argument_type > *t );

  /**
   * @todo write comments
   */
public:
  template< typename OPLUS >
  static dist_rose_tree< typename OPLUS::result_type > *
  dAcc( const OPLUS &oplus,
	const typename OPLUS::first_argument_type &unit_oplus,
	const dist_rose_tree< typename OPLUS::first_argument_type > *t );

  /**
   * @todo write comments
   */
public:
  template< typename OPLUS >
  static dist_rose_tree< typename OPLUS::result_type > *
  rAcc( const OPLUS &oplus,
	const typename OPLUS::first_argument_type &unit_oplus,
	const dist_rose_tree< typename OPLUS::first_argument_type > *t );

  /**
   * @todo write comments
   */
public:
  template< typename OPLUS >
  static dist_rose_tree< typename OPLUS::result_type > *
  lAcc( const OPLUS &oplus,
	const typename OPLUS::first_argument_type &unit_oplus,
	const dist_rose_tree< typename OPLUS::first_argument_type > *t );
  
};

#include "rose_tree_skeletons.tpp"

#endif // __ROSE_TREE_SKELETONS_H__
