/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file rose_tree_skeletons.tpp
 * @brief The implementation of class @a rose_tree_skeletons;
 *        a collection of skeletons on rose trees.
 * @author Kiminori MATSUZAKI
 */

#ifdef __ROSE_TREE_SKELETONS_H__
#ifndef __ROSE_TREE_SKELETONS_TPP__
#define __ROSE_TREE_SKELETONS_TPP__

class rose_tree_skeletons_impl
{
  friend class rose_tree_skeletons;
  
private:
  template< typename A, typename B >
  struct UnaryUndef : public skeleton::unary_function< A, B > {
    B operator()( const A& ) const {
      //      return static_cast< B >( 0 );
      return B( );
    }
  };

private:
  template< typename A, typename B, typename C >
  struct BinaryUndef : public skeleton::binary_function< A, B, C > {
    C operator()( const A&, const B& ) const {
      //      return static_cast< C >( );
      return C( );
    }
  };
  
private:
  template< typename A, typename B >
  struct UnaryConst : public skeleton::unary_function< A, B > {
    B result;
    UnaryConst( B result_ ) : result( result_ ) {}
    B operator()( const A& ) const { return result; }
  };

  template< typename A >
  struct UnaryId : public skeleton::unary_function< A, A > {
    A operator()( const A& x ) const { return x; }
  };

private:
  template< typename A, typename B, typename K >
  static dist_rose_tree< B > *
  map_adapter( const K &k,
	       const dist_rose_tree< A > *t );

private:
  template< typename A, typename B, typename C, typename K >
  static dist_rose_tree< C > *
  zipwith_adapter( const K &k,
		   const dist_rose_tree< A > *t1,
		   const dist_rose_tree< B > *t2 );

private:
  template< typename A, typename OPLUS, typename OTIMES >
  static A
  reduce_ring_adapter( const OPLUS &oplus,
		       const A& unit_oplus,
		       const OTIMES &otimes,
		       const A& unit_otimes,
		       const dist_rose_tree< A > *t );

private:
  template< typename A, typename B, typename OPLUS, typename OTIMES,
	    typename P1, typename P2, typename P3 >
  static B
  reduce_exdist_adapter( const OPLUS &oplus,
			 const A& unit_oplus,
			 const OTIMES &otimes,
			 const B& unit_otimes,
			 const P1 &p1,
			 const P2 &p2,
			 const P3 &p3,
			 const dist_rose_tree< A > *t );

private:
  template< typename A, typename OPLUS >
  static A
  reduce_assoc_adapter( const OPLUS oplus,
			const A& unit_oplus,
			const dist_rose_tree< A > *t );

private:
  template< typename A, typename OPLUS, typename OTIMES >
  static dist_rose_tree< A > *
  uAcc_ring_adapter( const OPLUS &oplus,
		     const A &unit_oplus,
		     const OTIMES &otimes,
		     const A &unit_otimes,
		     const dist_rose_tree< A > *t );

private:
  template< typename A, typename B, typename OPLUS, typename OTIMES,
	    typename P1, typename P2, typename P3 >
  static dist_rose_tree< B > *
  uAcc_exdist_adapter( const OPLUS &oplus,
		       const A &unit_oplus,
		       const OTIMES &otimes,
		       const B &unit_otimes,
		       const P1 &p1,
		       const P2 &p2,
		       const P3 &p3,
		       const dist_rose_tree< A > *t );

private:
  template< typename A, typename OPLUS >
  static dist_rose_tree< A > *
  uAcc_assoc_adapter( const OPLUS oplus,
		      const A& unit_oplus,
		      const dist_rose_tree< A > *t );

  template< typename A >
  struct reduce_ring_inter_t {
    A a, b, c;
    reduce_ring_inter_t( ) { }
    reduce_ring_inter_t( A a_, A b_, A c_ )
      : a( a_ ), b( b_ ), c( c_ ) { }
  };

  template< typename A, typename OPLUS, typename OTIMES >
  struct func_reduce_ring_k
    : skeleton::ternary_function< A, A, A, A > {
    const OPLUS &oplus;
    const OTIMES &otimes;
    func_reduce_ring_k( OPLUS oplus_, OTIMES otimes_ )
      : oplus( oplus_ ), otimes( otimes_ ) { }
    A operator()( const A& n, const A& l, const A& r ) const {
      return otimes( oplus( n, l ), r );
    }
  };

  template< typename A >
  struct func_reduce_ring_phi
    : skeleton::unary_function< A, reduce_ring_inter_t< A > > {
    const A &unit_oplus;
    const A &unit_otimes;
    func_reduce_ring_phi( A unit_oplus_, A unit_otimes_ )
      : unit_oplus( unit_oplus_ ), unit_otimes( unit_otimes_ ) { }
    reduce_ring_inter_t< A > operator()( const A& n ) const {
      return reduce_ring_inter_t< A >( n, unit_oplus, unit_otimes );
    }
  };

  template< typename A, typename OPLUS, typename OTIMES >
  struct func_reduce_ring_psiL
    : skeleton::ternary_function< reduce_ring_inter_t< A >,
				  A,
				  reduce_ring_inter_t< A >,
				  reduce_ring_inter_t< A > > {
    const OPLUS &oplus;
    const OTIMES &otimes;
    func_reduce_ring_psiL( OPLUS oplus_, OTIMES otimes_ )
      : oplus( oplus_ ), otimes( otimes_ ) { }
    reduce_ring_inter_t< A > operator()( const reduce_ring_inter_t< A > &n,
					 const A &l,
					 const reduce_ring_inter_t< A > &r ) const {
      return reduce_ring_inter_t< A >( oplus( n.b, r.a ),
				       oplus( n.b, r.b ),
				       otimes( otimes( oplus( n.a, l ),
						       oplus( n.b, r.c ) ),
					       n.c ) );
    }
  };

  template< typename A, typename OPLUS, typename OTIMES >
  struct func_reduce_ring_psiR
    : skeleton::ternary_function< reduce_ring_inter_t< A >,
				  A,
				  reduce_ring_inter_t< A >,
				  reduce_ring_inter_t< A > > {
    const OPLUS &oplus;
    const OTIMES &otimes;
    func_reduce_ring_psiR( OPLUS oplus_, OTIMES otimes_ )
      : oplus( oplus_ ), otimes( otimes_ ) { }
    reduce_ring_inter_t< A > operator()( const reduce_ring_inter_t< A > &n,
					 const A &r,
					 const reduce_ring_inter_t< A > &l ) const {
      return reduce_ring_inter_t< A >( oplus( n.a, l.a ),
				       oplus( n.a, l.b ),
				       otimes( otimes( oplus( n.a, l.c ),
						       oplus( n.b, r ) ),
					       n.c ) );
    }
  };

  template< typename A, typename OPLUS, typename OTIMES >
  struct func_reduce_ring_G
    : skeleton::ternary_function< reduce_ring_inter_t< A >,
				  A, A, A > {
    const OPLUS &oplus;
    const OTIMES &otimes;
    func_reduce_ring_G( OPLUS oplus_, OTIMES otimes_ )
      : oplus( oplus_ ), otimes( otimes_ ) { }
    A operator()( const reduce_ring_inter_t< A > &n,
		  const A &l,
		  const A &r ) const {
      return otimes( otimes( oplus( n.a, l ),
			     oplus( n.b, r ) ),
		     n.c );
    }
  };

  template< typename A, typename B >
  struct reduce_exdist_inter_t {
    A a;
    B b;
    A c;
    B d;
    reduce_exdist_inter_t( ) { }
    reduce_exdist_inter_t( A a_, B b_, A c_, B d_ )
      : a( a_ ), b( b_ ), c( c_ ), d( d_ ) { }
  };

  template< typename A, typename B, typename OPLUS, typename OTIMES >
  struct func_reduce_exdist_k
    : skeleton::ternary_function< A, B, B, B > {
    const OPLUS &oplus;
    const OTIMES &otimes;
    func_reduce_exdist_k( OPLUS oplus_, OTIMES otimes_ )
      : oplus( oplus_ ), otimes( otimes_ ) { }
    B operator()( const A& n, const B& l, const B& r ) const {
      return otimes( oplus( n, l ), r );
    }
  };
  
  template< typename A, typename B >
  struct func_reduce_exdist_phi
    : skeleton::unary_function< A, reduce_exdist_inter_t< A, B > > {
    const A& unit_oplus;
    const B& unit_otimes;
    func_reduce_exdist_phi( const A& unit_oplus_, const B& unit_otimes_ )
      : unit_oplus( unit_oplus_ ), unit_otimes( unit_otimes_ ) { }
    reduce_exdist_inter_t< A, B > operator()( const A& n ) const {
      return reduce_exdist_inter_t< A, B >( unit_oplus,
					    unit_otimes,
					    n,
					    unit_otimes );
    }
  };

  template< typename A, typename B, typename OPLUS, typename OTIMES,
	    typename P1, typename P2, typename P3 >
  struct func_reduce_exdist_psiL
    : skeleton::ternary_function< reduce_exdist_inter_t< A, B >,
				  B,
				  reduce_exdist_inter_t< A, B >,
				  reduce_exdist_inter_t< A, B > > {
    const OPLUS &oplus;
    const OTIMES &otimes;
    const P1 &p1;
    const P2 &p2;
    const P3 &p3;
    func_reduce_exdist_psiL( OPLUS oplus_, OTIMES otimes_,
			     P1 p1_, P2 p2_, P3 p3_ )
      : oplus( oplus_ ), otimes( otimes_ ), p1( p1_ ), p2( p2_ ), p3( p3_ ) { }
    reduce_exdist_inter_t< A, B > operator()( const reduce_exdist_inter_t< A, B > &n,
					      const B& l, 
					      const reduce_exdist_inter_t< A, B > &r ) const {
      B bb = otimes( n.b, oplus( n.c, l ) );
      return reduce_exdist_inter_t< A, B >( p1( n.a, bb, n.d, r.a, r.b, r.d ),
					    p2( n.a, bb, n.d, r.a, r.b, r.d ),
					    r.c,
					    p3( n.a, bb, n.d, r.a, r.b, r.d ) );
    }
  };

  template< typename A, typename B, typename OPLUS, typename OTIMES,
	    typename P1, typename P2, typename P3 >
  struct func_reduce_exdist_psiR
    : skeleton::ternary_function< reduce_exdist_inter_t< A, B >,
				  B,
				  reduce_exdist_inter_t< A, B >,
				  reduce_exdist_inter_t< A, B > > {
    const OPLUS &oplus;
    const OTIMES &otimes;
    const B& unit_otimes;
    const P1 &p1;
    const P2 &p2;
    const P3 &p3;
    func_reduce_exdist_psiR( OPLUS oplus_, OTIMES otimes_, const B& unit_otimes_, 
			     P1 p1_, P2 p2_, P3 p3_ )
      : oplus( oplus_ ), otimes( otimes_ ), unit_otimes( unit_otimes_ ),
	p1( p1_ ), p2( p2_ ), p3( p3_ ) { }
    reduce_exdist_inter_t< A, B > operator()( const reduce_exdist_inter_t< A, B > &n,
					      const B& r, 
					      const reduce_exdist_inter_t< A, B > &l ) const {
      B rd = otimes( r, n.d );
      A t1 = p1( n.c, unit_otimes, unit_otimes, l.a, l.b, l.d );
      B t2 = p2( n.c, unit_otimes, unit_otimes, l.a, l.b, l.d );
      B t3 = p3( n.c, unit_otimes, unit_otimes, l.a, l.b, l.d );
      return reduce_exdist_inter_t< A, B >( p1( n.a, n.b, rd, t1, t2, t3 ),
					    p2( n.a, n.b, rd, t1, t2, t3 ),
					    l.c,
					    p3( n.a, n.b, rd, t1, t2, t3 ) );
    }
  };

  template< typename A, typename B, typename OPLUS, typename OTIMES >
  struct func_reduce_exdist_G
    : skeleton::ternary_function< reduce_exdist_inter_t< A, B >, B, B, B > {
    const OPLUS &oplus;
    const OTIMES &otimes;
    func_reduce_exdist_G( OPLUS oplus_, OTIMES otimes_ )
      : oplus( oplus_ ), otimes( otimes_ ) { }
    B operator()( const reduce_exdist_inter_t< A, B >& n,
		  const B& l,
		  const B& r ) const {
      return oplus( n.a,
		    otimes( otimes( otimes( n.b, 
					    oplus( n.c, l ) ),
				    r ),
			    n.d ) );
    }
  };

  template< typename A, typename OPLUS >
  struct func_uAcc_assoc_ternary
    : skeleton::ternary_function< A, A, A, A > {
    const OPLUS oplus;
    func_uAcc_assoc_ternary( OPLUS oplus_ ) : oplus( oplus_ ) { }
    A operator()( const A& n, const A& x, const A& m ) const {
      return oplus( n, oplus( x, m ) );
    }
  };
    
private:
  template< typename A, typename OPLUS >
  static dist_rose_tree< A > *
  dAcc_adapter( const OPLUS oplus,
		const A& unit_oplus,
		const dist_rose_tree< A > *t );
  
private:
  template< typename A, typename OPLUS >
  static dist_rose_tree< A > *
  rAcc_adapter( const OPLUS& oplus,
		const A& unit_oplus,
		const dist_rose_tree< A > *t );

  template< typename A >
  struct rAcc_inter_struct_t {
    bool isunit;
    bool p;
    A a;
    A b;
    rAcc_inter_struct_t( ) { }
    rAcc_inter_struct_t( bool p_, const A& a_, const A& b_ )
      : isunit( false ), p( p_ ), a( a_ ), b( b_ ) { }
    static rAcc_inter_struct_t< A > unit( const A& u ) {
      rAcc_inter_struct_t< A > ret;
      ret.isunit = true; ret.p = true; ret.a = u; ret.b = u;
      return ret;
    }
  };

  template < typename A, typename OPLUS >
  struct func_rAcc_otimes
    : public skeleton::binary_function< rAcc_inter_struct_t< A >,
					rAcc_inter_struct_t< A >,
					rAcc_inter_struct_t< A > > {
    const OPLUS& oplus;
    func_rAcc_otimes( const OPLUS &oplus_ ) : oplus( oplus_ ) { }
    rAcc_inter_struct_t< A > operator()( const rAcc_inter_struct_t< A >& lhs,
					 const rAcc_inter_struct_t< A >& rhs ) const {
      if ( lhs.isunit ) return rhs;
      if ( rhs.isunit ) return lhs;
      return rAcc_inter_struct_t< A >( lhs.p && rhs.p,
				       oplus( lhs.a, rhs.a ),
				       rhs.p ? oplus( lhs.b, rhs.a ) : rhs.b );
    }
  };

  template < typename A >
  static rAcc_inter_struct_t< A > unit_func_rAcc_otimes( const A& unit )
  {
    return rAcc_inter_struct_t< A >::unit( unit );
    // return rAcc_inter_struct_t< A >( true, unit, static_cast< A >( 0 ) );
  }
    
  template < typename A >
  struct func_rAcc_gl
    : public skeleton::unary_function< A, rAcc_inter_struct_t< A > > {
    const A& unit;
    func_rAcc_gl( const A& unit_ ) : unit( unit_ ) { }
    rAcc_inter_struct_t< A > operator()( const A& ) const {
      return rAcc_inter_struct_t< A >( false, static_cast< A >( 0 ), unit );
    }
  };

  template < typename A >
  struct func_rAcc_gr
    : public skeleton::unary_function< A, rAcc_inter_struct_t< A > > {
    rAcc_inter_struct_t< A > operator()( const A& x ) const {
      return rAcc_inter_struct_t< A >( true, x, static_cast< A >( 0 ) );
    }
  };

  template< typename A > 
  struct func_rAcc_k
    : public skeleton::unary_function< rAcc_inter_struct_t< A >, A > {
    A operator()( const rAcc_inter_struct_t< A >& x ) const {
      return x.p ? x.a : x.b;
    }
  };

private:
  template< typename A, typename OPLUS >
  static dist_rose_tree< A > *
  lAcc_adapter( const OPLUS& oplus,
		const A& unit_oplus,
		const dist_rose_tree< A > *t );

  template < typename A >
  struct lAcc_inter_struct_t {
    bool p;
    A a;
    A b;
    lAcc_inter_struct_t( ) { }
    lAcc_inter_struct_t( bool p_, const A& a_, const A& b_ )
      : p( p_ ), a( a_ ), b( b_ ) { }
  };
  
  template < typename A, typename OPLUS >
  struct func_lAcc_k : public skeleton::ternary_function< A, A, A, A > {
    const OPLUS &oplus;
    func_lAcc_k( OPLUS oplus_ ) : oplus( oplus_ ) {}
    A operator()( const A& n, const A& /* l */, const A& r ) const {
      return oplus( n, r );
    }
  };

  template < typename A >
  struct func_lAcc_phi : public skeleton::unary_function< A, lAcc_inter_struct_t< A > > {
    lAcc_inter_struct_t< A > operator()( const A& n ) const {
      return lAcc_inter_struct_t< A >( true, n, static_cast< A >( 0 ) );
    }
  };

  template < typename A, typename OPLUS >
  struct func_lAcc_psiL
    : public skeleton::ternary_function< lAcc_inter_struct_t< A >,
					 A,
					 lAcc_inter_struct_t< A >,
					 lAcc_inter_struct_t< A > > {
    const OPLUS &oplus;
    func_lAcc_psiL( OPLUS oplus_ ) : oplus( oplus_ ) {};
    lAcc_inter_struct_t< A > operator()( const lAcc_inter_struct_t< A >& arg1,
				       const A& /* l */,
				       lAcc_inter_struct_t< A >& arg2 ) const {
      return lAcc_inter_struct_t< A >( arg1.p && arg2.p,
				     oplus( arg1.a, arg2.a ),
				     arg1.p ? ( oplus( arg1.a, arg2.b ) ) : arg1.b );
    }
  };

  template < typename A, typename OPLUS >
  struct func_lAcc_psiR
    : public skeleton::ternary_function< lAcc_inter_struct_t< A >,
					 A,
					 lAcc_inter_struct_t< A >,
					 lAcc_inter_struct_t< A > > {
    const OPLUS &oplus;
    func_lAcc_psiR( OPLUS oplus_ ) : oplus( oplus_ ) {};
    lAcc_inter_struct_t< A > operator()( const lAcc_inter_struct_t< A >& arg1,
				       const A& r,
				       lAcc_inter_struct_t< A >& /* arg2 */ ) const {
      return lAcc_inter_struct_t< A >( false, 
				     static_cast< A >( 0 ), 
				     arg1.p ? ( oplus( arg1.a, r ) ) : arg1.b );
    }
  };

  template < typename A, typename OPLUS >
  struct func_lAcc_G
    : public skeleton::ternary_function< lAcc_inter_struct_t< A >, A, A, A > {
    const OPLUS &oplus;
    func_lAcc_G( OPLUS oplus_ ) : oplus( oplus_ ) {};
    A operator()( const lAcc_inter_struct_t< A >& arg,
		  const A& /* l */,
		  const A& r ) const {
      return arg.p ? oplus( arg.a, r ) : arg.b;
    }
  };
};

/* private, static */
template< typename A, typename B, typename K >
dist_rose_tree< B > *
rose_tree_skeletons_impl::map_adapter( const K &k,
					const dist_rose_tree< A > *t )
{
  dist_tree< B > *bt
    = tree_skeletons::map( UnaryUndef< A, B >( ), k, t->btree );
  return new dist_rose_tree< B >( bt );
}

/* private, static */
template< typename A, typename B, typename C, typename K >
dist_rose_tree< C > *
rose_tree_skeletons_impl::zipwith_adapter( const K &k,
					   const dist_rose_tree< A > *t1,
					   const dist_rose_tree< B > *t2 )
{
  dist_tree< C > *ct
    = tree_skeletons::zipwith( BinaryUndef< A, B, C >( ), k, t1->btree, t2->btree );
  return new dist_rose_tree< C >( ct );
}

/* private, static */
template< typename A, typename OPLUS, typename OTIMES >
A
rose_tree_skeletons_impl::reduce_ring_adapter( const OPLUS &oplus,
					       const A& unit_oplus,
					       const OTIMES &otimes,
					       const A& unit_otimes,
					       const dist_rose_tree< A > *t )
{
  return tree_skeletons::reduce( UnaryConst< A, A >( unit_otimes ),
				 func_reduce_ring_k< A, OPLUS, OTIMES >( oplus, otimes ),
				 func_reduce_ring_phi< A >( unit_oplus, unit_otimes ),
				 func_reduce_ring_psiL< A, OPLUS, OTIMES >( oplus, otimes ),
				 func_reduce_ring_psiR< A, OPLUS, OTIMES >( oplus, otimes ),
				 func_reduce_ring_G< A, OPLUS, OTIMES >( oplus, otimes ),
				 t->btree );
}

/* private, static */
template< typename A, typename B, typename OPLUS, typename OTIMES,
	  typename P1, typename P2, typename P3 >
B
rose_tree_skeletons_impl::reduce_exdist_adapter( const OPLUS &oplus,
						 const A& unit_oplus,
						 const OTIMES &otimes,
						 const B& unit_otimes,
						 const P1 &p1,
						 const P2 &p2,
						 const P3 &p3,
						 const dist_rose_tree< A > *t )
{
  return tree_skeletons::reduce( UnaryConst< A, B >( unit_otimes ),
				 func_reduce_exdist_k< A, B, OPLUS, OTIMES >( oplus, otimes ),
				 func_reduce_exdist_phi< A, B >( unit_oplus, unit_otimes ),
				 func_reduce_exdist_psiL< A, B, OPLUS, OTIMES, P1, P2, P3 >( oplus, otimes, p1, p2, p3 ),
				 func_reduce_exdist_psiR< A, B, OPLUS, OTIMES, P1, P2, P3 >( oplus, otimes, unit_otimes, p1, p2, p3 ),
				 func_reduce_exdist_G< A, B, OPLUS, OTIMES >( oplus, otimes ), 
				 t->btree );
}

/* private, static */
template< typename A, typename OPLUS >
A
rose_tree_skeletons_impl::reduce_assoc_adapter( const OPLUS oplus,
						const A& unit_oplus,
						const dist_rose_tree< A > *t )
{
  return tree_skeletons::reduce( UnaryConst< A, A >( unit_oplus ),
				 func_uAcc_assoc_ternary< A, OPLUS >( oplus ),
				 UnaryId< A >( ),
				 func_uAcc_assoc_ternary< A, OPLUS >( oplus ),
				 func_uAcc_assoc_ternary< A, OPLUS >( oplus ),
				 func_uAcc_assoc_ternary< A, OPLUS >( oplus ),
				 t->btree );
}

/* private, static */
template< typename A, typename OPLUS, typename OTIMES >
dist_rose_tree< A > *
rose_tree_skeletons_impl::uAcc_ring_adapter( const OPLUS &oplus,
					     const A &unit_oplus,
					     const OTIMES &otimes,
					     const A &unit_otimes,
					     const dist_rose_tree< A > *t )
{
  dist_tree< A > *bt1
    = tree_skeletons::uAcc( UnaryConst< A, A >( unit_otimes ),
			    func_reduce_ring_k< A, OPLUS, OTIMES >( oplus, otimes ),
			    func_reduce_ring_phi< A >( unit_oplus, unit_otimes ),
			    func_reduce_ring_psiL< A, OPLUS, OTIMES >( oplus, otimes ),
			    func_reduce_ring_psiR< A, OPLUS, OTIMES >( oplus, otimes ),
			    func_reduce_ring_G< A, OPLUS, OTIMES >( oplus, otimes ),
			    t->btree );
  dist_tree< A > *bt2
    = tree_skeletons::shiftl( unit_otimes, bt1 );
  if ( bt1 ) delete bt1;

  dist_tree< A > *bt3
    = tree_skeletons::zipwith( BinaryUndef< A, A, A >( ), oplus,
			       t->btree, bt2 );
  if ( bt2 ) delete bt2;

  return new dist_rose_tree< A >( bt3 );
}


/* private, static */
template< typename A, typename B, typename OPLUS, typename OTIMES,
	  typename P1, typename P2, typename P3 >
dist_rose_tree< B > *
rose_tree_skeletons_impl::uAcc_exdist_adapter( const OPLUS &oplus,
					       const A &unit_oplus,
					       const OTIMES &otimes,
					       const B &unit_otimes,
					       const P1 &p1,
					       const P2 &p2,
					       const P3 &p3,
					       const dist_rose_tree< A > *t )
{
  dist_tree< B > *bt1
    = tree_skeletons::uAcc( UnaryConst< A, B >( unit_otimes ),
			    func_reduce_exdist_k< A, B, OPLUS, OTIMES >( oplus, otimes ),
			    func_reduce_exdist_phi< A, B >( unit_oplus, unit_otimes ),
			    func_reduce_exdist_psiL< A, B, OPLUS, OTIMES, P1, P2, P3 >( oplus, otimes, p1, p2, p3 ),
			    func_reduce_exdist_psiR< A, B, OPLUS, OTIMES, P1, P2, P3 >( oplus, otimes, unit_otimes, p1, p2, p3 ),
			    func_reduce_exdist_G< A, B, OPLUS, OTIMES >( oplus, otimes ), 
			    t->btree );
  dist_tree< B > *bt2
    = tree_skeletons::shiftl( unit_otimes, bt1 );
  if ( bt1 ) delete bt1;

  dist_tree< B > *bt3
    = tree_skeletons::zipwith( BinaryUndef< A, B, B >( ), oplus,
			       t->btree, bt2 );
  if ( bt2 ) delete bt2;

  return new dist_rose_tree< B >( bt3 );
}

/* private, static */
template< typename A, typename OPLUS >
dist_rose_tree< A > *
rose_tree_skeletons_impl::uAcc_assoc_adapter( const OPLUS oplus,
					      const A& unit_oplus,
					      const dist_rose_tree< A > *t )
{
  dist_tree< A > *bt1
    = tree_skeletons::uAcc( UnaryConst< A, A >( unit_oplus ),
			    func_uAcc_assoc_ternary< A, OPLUS >( oplus ),
			    UnaryId< A >( ),
			    func_uAcc_assoc_ternary< A, OPLUS >( oplus ),
			    func_uAcc_assoc_ternary< A, OPLUS >( oplus ),
			    func_uAcc_assoc_ternary< A, OPLUS >( oplus ),
			    t->btree );
  dist_tree< A > *bt2
    = tree_skeletons::shiftl( unit_oplus, bt1 );
  if ( bt1 ) delete bt1;

  dist_tree< A > *bt3
    = tree_skeletons::zipwith( BinaryUndef< A, A, A >( ), oplus,
			       t->btree, bt2 );
  if ( bt2 ) delete bt2;

  return new dist_rose_tree< A >( bt3 );
}

/* private, static */
template< typename A, typename OPLUS >
dist_rose_tree< A > *
rose_tree_skeletons_impl::dAcc_adapter( const OPLUS oplus,
					const A& unit_oplus,
					const dist_rose_tree< A > *t )
{
  dist_tree< A > *bt
    = tree_skeletons::dAcc( oplus, unit_oplus,
			    UnaryId< A >( ), UnaryConst< A, A >( unit_oplus ),
			    unit_oplus, t->btree );
  return new dist_rose_tree< A >( bt );
}

/* private, static */
template< typename A, typename OPLUS >
dist_rose_tree< A > *
rose_tree_skeletons_impl::rAcc_adapter( const OPLUS& oplus,
					const A& unit_oplus,
					const dist_rose_tree< A > *t )
{
  dist_tree< rAcc_inter_struct_t< A > > *bt1
    = tree_skeletons::dAcc( func_rAcc_otimes< A, OPLUS >( oplus ),
			    unit_func_rAcc_otimes< A >( unit_oplus ),
			    func_rAcc_gl< A >( unit_oplus ),
			    func_rAcc_gr< A >( ),
			    unit_func_rAcc_otimes< A >( unit_oplus ),
			    t->btree );
  dist_tree< A > *bt2
    = tree_skeletons::map( UnaryUndef< rAcc_inter_struct_t< A >, A >( ),
			   func_rAcc_k< A >( ),
			   bt1 );

  if ( bt1 ) delete bt1;
  return new dist_rose_tree< A >( bt2 );
}

/* private, static */
template< typename A, typename OPLUS >
dist_rose_tree< A > *
rose_tree_skeletons_impl::lAcc_adapter( const OPLUS& oplus,
					const A& unit_oplus,
					const dist_rose_tree< A > *t )
{
  dist_tree< A > *bt1
    = tree_skeletons::uAcc( UnaryConst< A, A >( unit_oplus ),
			    func_lAcc_k< A, OPLUS >( oplus ),
			    func_lAcc_phi< A >( ),
			    func_lAcc_psiL< A, OPLUS >( oplus ),
			    func_lAcc_psiR< A, OPLUS >( oplus ),
			    func_lAcc_G< A, OPLUS >( oplus ),
			    t->btree );
  dist_tree< A > *bt2
    = tree_skeletons::shiftr( unit_oplus, bt1 );

  if ( bt1 ) delete bt1;
  return new dist_rose_tree< A >( bt2 );
}


////////////////////////////////////////////////////////////////////////
// functions for calling the adapter function.

/* public, static */
template< typename K >
dist_rose_tree< typename K::result_type > *
rose_tree_skeletons::map( const K &k,
			  const dist_rose_tree< typename K::argument_type > *t )
{
  return rose_tree_skeletons_impl::map_adapter
    < typename K::argument_type, typename K::result_type, K >
    ( k, t );
}

/* public, static */
template< typename K >
dist_rose_tree< typename K::result_type > *
rose_tree_skeletons::zipwith( const K &k,
			      const dist_rose_tree< typename K::first_argument_type > *t1,
			      const dist_rose_tree< typename K::second_argument_type > *t2) {
  return rose_tree_skeletons_impl::zipwith_adapter
    < typename K::first_argument_type, typename K::second_argument_type, typename K::result_type, K >
    ( k, t1, t2 );
}

/* public, static */
template< typename OPLUS, typename OTIMES >
typename OTIMES::result_type
rose_tree_skeletons::reduce_ring( const OPLUS &oplus,
				  const typename OPLUS::first_argument_type &unit_oplus,
				  const OTIMES &otimes,
				  const typename OTIMES::first_argument_type &unit_otimes,
				  const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::reduce_ring_adapter
    < typename OPLUS::first_argument_type, OPLUS, OTIMES >
    ( oplus, unit_oplus, otimes, unit_otimes, t );
}

/* public, static */
template< typename OPLUS, typename OTIMES,
	  typename P1, typename P2, typename P3 >
typename OPLUS::result_type
rose_tree_skeletons::reduce_exdist( const OPLUS &oplus,
				    const typename OPLUS::first_argument_type &unit_oplus,
				    const OTIMES &otimes,
				    const typename OTIMES::first_argument_type &unit_otimes,
				    const P1 &p1,
				    const P2 &p2,
				    const P3 &p3,
				    const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::reduce_exdist_adapter
    < typename OPLUS::first_argument_type, typename OTIMES::first_argument_type,
    OPLUS, OTIMES, P1, P2, P3 >
    ( oplus, unit_oplus, otimes, unit_otimes, p1, p2, p3, t );
}

/* public, static */
template< typename OPLUS >
typename OPLUS::result_type
rose_tree_skeletons::reduce_assoc( const OPLUS &oplus,
				   const typename OPLUS::first_argument_type &unit_oplus,
				   const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::reduce_assoc_adapter
    < typename OPLUS::first_argument_type, OPLUS >
    ( oplus, unit_oplus, t );
}

/* public, static */
template< typename OPLUS, typename OTIMES >
dist_rose_tree< typename OTIMES::result_type > *
rose_tree_skeletons::uAcc_ring( const OPLUS &oplus,
				const typename OPLUS::first_argument_type &unit_oplus,
				const OTIMES &otimes,
				const typename OTIMES::first_argument_type &unit_otimes,
				const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::uAcc_ring_adapter
    < typename OPLUS::first_argument_type, OPLUS, OTIMES >
    ( oplus, unit_oplus, otimes, unit_otimes, t );
}

/* public, static */
template< typename OPLUS, typename OTIMES,
	  typename P1, typename P2, typename P3 >
dist_rose_tree< typename OPLUS::result_type > *
rose_tree_skeletons::uAcc_exdist( const OPLUS &oplus,
				  const typename OPLUS::first_argument_type &unit_oplus,
				  const OTIMES &otimes,
				  const typename OTIMES::first_argument_type &unit_otimes,
				  const P1 &p1,
				  const P2 &p2,
				  const P3 &p3,
				  const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::uAcc_exdist_adapter
    < typename OPLUS::first_argument_type, typename OPLUS::result_type,
    OPLUS, OTIMES, P1, P2, P3 >
    ( oplus, unit_oplus, otimes, unit_otimes, p1, p2, p3, t );
}

/* public, static */
template< typename OPLUS >
dist_rose_tree< typename OPLUS::result_type > *
rose_tree_skeletons::uAcc_assoc( const OPLUS &oplus,
				 const typename OPLUS::first_argument_type &unit_oplus,
				 const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::uAcc_assoc_adapter
    < typename OPLUS::first_argument_type, OPLUS >
    ( oplus, unit_oplus, t );
}

/* public, static */
template< typename OPLUS >
dist_rose_tree< typename OPLUS::result_type > *
rose_tree_skeletons::dAcc( const OPLUS &oplus,
			   const typename OPLUS::first_argument_type &unit_oplus,
			   const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::dAcc_adapter
    < typename OPLUS::first_argument_type, OPLUS >
    ( oplus, unit_oplus, t );
}

/* public, static */
template< typename OPLUS >
dist_rose_tree< typename OPLUS::result_type > *
rose_tree_skeletons::rAcc( const OPLUS &oplus,
			   const typename OPLUS::first_argument_type &unit_oplus,
			   const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::rAcc_adapter
    < typename OPLUS::first_argument_type, OPLUS >
    ( oplus, unit_oplus, t );
}

/* public, static */
template< typename OPLUS >
dist_rose_tree< typename OPLUS::result_type > *
rose_tree_skeletons::lAcc( const OPLUS &oplus,
			   const typename OPLUS::first_argument_type &unit_oplus,
			   const dist_rose_tree< typename OPLUS::first_argument_type > *t )
{
  return rose_tree_skeletons_impl::lAcc_adapter
    < typename OPLUS::first_argument_type, OPLUS >
    ( oplus, unit_oplus, t );
}

#endif /* __ROSE_TREE_SKELETONS_TPP__ */
#endif /* __ROSE_TREE_SKELETONS_H__ */
