/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file node.h
 * @brief The header file of class @a node. 
 * @author Kiminori MATSUZAKI
 */

#ifndef __NODE_H__
#define __NODE_H__

/**
 * @brief A template class for a node of a tree. 
 */
template< typename A >
class node {
  ////////////////////////////////////////////////////////////////////////
  // constants
public:
  typedef enum {
    /** @brief For a normal node or a leaf. */
    NORMAL = 0,
    /** @brief For a cutting node. */
    CUTNODE = 1, 
  } nodetype_t;

  /**
   * @brief Enumerations for flags in serialized array, and in file.
   */
public:
  typedef enum {
    /** @brief Internal node. */
    SER_NODE = 0,
    /** @brief Leaf */
    SER_LEAF = 1,
    /** @brief Cut-Node */
    SER_CUTNODE = 2,
  } serializetype_t;
    
  ////////////////////////////////////////////////////////////////////////
  // member variables
public:
  /** @brief The type of this node. Usually, @c NORMAL. */
  nodetype_t nodetype;

  /** @brief The pointer to the left-child node. */
  node< A > *left;
  /** @brief The pointer to the right-child node. */
  node< A > *right;
  /** @brief The pointer to the parent node. */
  node< A > *parent;
  
  /** @brief The value of this node. */
  A   value;

  ////////////////////////////////////////////////////////////////////////
  // Constructors and Destructor
public:
  /**
   * @brief Construct a normal node by specifying @c value, and @c nodetype. 
   */
  node( const A& value, nodetype_t nodetype = NORMAL );

  /**
   * @brief Construct a node without the value
   */
  node( );
  
  /**
   * @brief Destruct a node only.
   * @attention This destructor does not recursively free the memory.
   */
  ~node( );

  ////////////////////////////////////////////////////////////////////////
  // Access Method

public:
  /**
   * @retval true if the node is a leaf, i.e., it has no children. 
   * @retval false if it has children.
   */
  bool isLeaf( ) const;

};

#include "node.tpp"

#endif /* __NODE_H__ */
