/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file test_dist_tree.cpp
 * @brief Test code for dist_tree.h and dist_tree.tpp
 * @author Kiminori MATSUZAKI
 */

#include <assert.h>
#include <iostream>

#include <skeleton.h>
#include <dist_tree.h>

using namespace std;

#define DEBUG_OUT 0

#if DEBUG_OUT
#define debug( x )   x
#else
#define debug( x )
#endif /* DEBUG_OUT */

void
test_dist_tree_read_from_file_and_destructor_and_print( )
{
  const char filename[] = "dist_tree_data.txt";
  dist_tree< double >* tree
    = dist_tree< double >::read_from_file( filename );

  debug( tree->print( ) );
  if ( tree ) delete tree;
}

void
test_dist_tree_read_from_file_and_write_to_file( )
{
  const char filename_in[]  = "dist_tree_data.txt";
  const char filename_out[] = "result_test_dist_tree_read_from_file_and_write.txt";
    
  dist_tree< double >* tree
    = dist_tree< double >::read_from_file( filename_in );
      
  assert( tree->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr
      << "Check the output file below is the same as input file:" << std::endl
      << "Input:  " << filename_in  << std::endl
      << "Output: " << filename_out << std::endl;
  }
  
  if ( tree ) delete tree;
}

int SketoMain( int, char *[] )
{
  skeleton::cout << "Testing " __FILE__ << endl;

  test_dist_tree_read_from_file_and_destructor_and_print( );
  test_dist_tree_read_from_file_and_write_to_file( );
  
  skeleton::cout << "   ---> OK " __FILE__ << endl;

  return 0;
}
