/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file test_tree_util.cpp
 * @brief Test code for tree_util.h and tree_util.tpp
 * @author Kiminori MATSUZAKI
 */

#include <assert.h>
#include <iostream>

#include <skeleton.h>
#include <dist_tree.h>
#include <tree_util.h>

using namespace std;

void test_tree_util_read_from_file_and_free_all_tree( )
{
  if ( skeleton::rank != 0 ) return;

  const char filename[] = "seq_tree_data_small.txt";
  node< char >* tree = tree_util::read_tree_from_file< char >( filename );

  assert( tree );
  assert( tree->value == 'A' );
  assert( tree->left->value == 'B' );
  assert( tree->left->left->value == 'C' );
  assert( tree->left->right->value == 'D' );
  assert( tree->right->value == 'E' );

  assert( tree->left->parent == tree );
  assert( tree->left->left->parent == tree->left );
  assert( tree->left->right->parent == tree->left );
  assert( tree->right->parent == tree );

  tree_util::free_all_tree( tree );
}

void test_tree_util_distribute_and_write( )
{
  const char filename_in[]  = "seq_tree_data.txt";
  const char filename_out[] = "result_test_tree_util_distributed_and_write.txt";

  node< double >* tree
    = tree_util::read_tree_from_file< double >( filename_in );
  
  if ( !tree_util::dist_and_write_to_file( tree,
					   skeleton::procs, 
					   filename_out ) ) {
    std::cerr << "Error in shapring dist_tree. " << std::endl;
    return;
  }

  std::cout << "Check the distributed tree in:" << std::endl
	    << filename_out << std::endl;

  tree_util::free_all_tree( tree );
}


int SketoMain( int, char** )
{
  if ( skeleton::rank != 0 ) return 0;

  cout << "Testing " __FILE__ << endl;
  
  test_tree_util_read_from_file_and_free_all_tree( );
  test_tree_util_distribute_and_write( );
  
  cout << "   ---> OK " __FILE__ << endl;
  return 0;
}
