/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file functions.h
 * @brief The header for base classes of function objects.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#ifndef __FUNCTIONS_H__
#define __FUNCTIONS_H__

namespace skeleton {
  template< typename Arg, typename Res >
  struct unary_function {
    typedef Arg argument_type;
    typedef Res result_type;
  };
  
  template< typename Arg, typename Arg2>
  struct unary_ptr_function {
    typedef Arg  first_argument_type;
    typedef Arg2 second_argument_type;
  };
  
  template< typename Arg, typename Arg2, typename Res >
  struct binary_function {
    typedef Arg  first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Res  result_type;
  };
  
  template< typename Arg, typename Arg2, typename Arg3 >
  struct binary_ptr_function {
    typedef Arg  first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Arg3 third_argument_type;
  };
  
  template< typename Arg1, typename Arg2, typename Arg3, typename Res >
  struct ternary_function {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Arg3 third_argument_type;
    typedef Res  result_type;
  };
  
  template< typename Arg >
  struct Id : public unary_function< Arg, Arg > {
    Arg operator()( Arg x ) const { return x; }
  };
  
  template< typename Arg >
  struct Snd : public binary_function< Arg, Arg, Arg > {
    Arg operator()( Arg a, Arg b ) const { return b; }
  };
  
  template< typename F>
  inline Id< typename F::result_type >
  make_id( const F &f )
  {
    return Id< typename F::result_type >();
  }
  
  template< typename F >
  inline Snd< typename F::result_type >
  make_snd( const F &f )
  {
    return Snd< typename F::result_type >();
  }
}

#endif // __FUNCTIONS_H__
