/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file produce.h
 * @brief The header for produce functions.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#ifndef __PRODUCE_H__
#define __PRODUCE_H__

#include <utility>
#include "functions.h"

namespace skeleton {
  template< typename F, typename G >
  class product_function
    : public skeleton::binary_function< typename F::argument_type,
					typename G::argument_type,
					std::pair< typename F::result_type, typename G::result_type > >
  {
    F f;
    G g;
  public:
    product_function( const F &f_, const G &g_ ) : f( f_ ), g( g_ )
    { }
    std::pair< typename F::result_type, typename G::result_type >
    operator()( typename F::argument_type x, typename G::argument_type y ) const
    {
      return std::make_pair( f( x ), g( y ) );
    }
  };
  
  template< typename F, typename G >
  class product_ptr_function
    : public skeleton::binary_ptr_function< std::pair< typename F::first_argument_type, typename G::first_argument_type >,
					    typename F::second_argument_type,
					    typename G::second_argument_type >
  {
    F f;
    G g;
  public:
    product_ptr_function( const F &f_, const G &g_ ) : f( f_ ), g( g_ )
    { }
    void operator()( std::pair< typename F::first_argument_type, typename G::fist_argument_type > *r,
		     const typename F::second_argument_type *x,
		     const typename G::second_argument_type *y ) const
    {
      typename F::first_argument_type a;
      typename G::first_argument_type b;
      f( &a, x );
      g( &b, y );
      *r = std::make_pair(a, b);
    }
  };
  
  template< typename F, typename G >
  inline skeleton::product_function< F, G >
  produce_functions( const F &f, const G &g )
  {
    return skeleton::product_function< F, G >( f, g );
  }
  
  template< typename F, typename G >
  inline skeleton::product_ptr_function< F, G >
  produce_ptr_functions( const F &f, const G &g )
  {
    return skeleton::product_ptr_function< F, G >( f, g );
  }
}

#endif // __PRODUCE_H__
