/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file resolve.h
 * @brief The header for resolve functions.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#ifndef __RESOLVE_H__

#define __RESOLVE_H__

#include <utility>
#include "functions.h"

namespace skeleton {
  template< typename F >
  class resolving_function
    : public skeleton::binary_function< typename F::argument_type::first_type,
					typename F::argument_type::second_type,
					typename F::result_type >
  {
    F f;
  public:
    resolving_function( const F &f_ ) : f( f_ )
    { }
    typename F::result_type
    operator()( typename F::argument_type::first_type x,
		typename F::argument_type::second_type y ) const
    {
      return f( std::make_pair( x, y ) );
    }
  };
  
  template< typename F >
  class resolving_ptr_function
    : public skeleton::binary_ptr_function< typename F::first_argument_type,
					    typename F::second_argument_type::first_type,
					    typename F::second_argument_type::second_type >
  {
    F f;
  public:
    resolving_ptr_function( const F &f_ ) : f( f_ )
    { }
    void
    operator()( typename F::first_argument_type *r,
		const typename F::second_argument_type::first_type *x,
		const typename F::second_argument_type::second_type *y ) const
    {
      f( r, std::make_pair( x, y ) );
    }
  };
  
  template< typename F >
  inline skeleton::resolving_function< F >
  resolve_function( const F &f )
  {
    return skeleton::resolving_function< F >( f );
  }
  
  template< typename F >
  inline skeleton::resolving_ptr_function< F >
  resolve_ptr_function( const F &f )
  {
    return skeleton::resolving_ptr_function< F >( f );
  }
}

#endif // __RESOLVE_H__
