/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file template_mpitype.h
 * @brief The header for template mpitype on distributed array.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 * @author Kiminori MATSUZAKI (kmatsu@ipl.t.u-tokyo.ac.jp)
 */

#ifndef __TEMPLATE_MPITYPE_H__
#define __TEMPLATE_MPITYPE_H__
#include <utility>

template<class A>
class template_mpitype {
public:
  static const MPI_Datatype type() { return A::mpitype; }
};

template<>
class template_mpitype<char> {
public:
  static const MPI_Datatype type() { return MPI_CHAR; }
};

template<>
class template_mpitype<short> {
public:
  static const MPI_Datatype type( ) { return MPI_SHORT; }
};

template<>
class template_mpitype<int> {
public:
  static const MPI_Datatype type() { return MPI_INT; }
};

template<>
class template_mpitype<float> {
public:
  static const MPI_Datatype type() { return MPI_FLOAT; }
};

template<>
class template_mpitype<double> {
public:
  static const MPI_Datatype type() { return MPI_DOUBLE; }
};

template<>
class template_mpitype<unsigned int> {
public:
  static const MPI_Datatype type() { return MPI_UNSIGNED; }
};

template<typename A, typename B>
class template_mpitype< std::pair<A, B> > {
public:
  static const MPI_Datatype type()
  {
    std::pair<A, B> temp;
    int blength_list[2] = {1, 1};
    MPI_Datatype datatype_list[2] = {template_mpitype<A>::type(),
				     template_mpitype<B>::type()};
    MPI_Aint displacement_list[2];
    MPI_Aint start_address, address;
    MPI_Address(&temp, &start_address);

    MPI_Address(&temp.first, &address);
    displacement_list[0] = address - start_address;

    MPI_Address(&temp.second, &address);
    displacement_list[1] = address - start_address;

    MPI_Datatype newtype;
    MPI_Type_struct(2, blength_list, displacement_list,
		     datatype_list, &newtype);
    MPI_Type_commit(&newtype);

    return newtype;
  }
};

#endif // __TEMPLATE_MPITYPE_H__
