/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file line_of_sight.cpp
 * @brief This program search lines which you can see.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#include <cstdlib>
#include <cfloat>
#include <ctime>
#include "list_skeletons.h"

struct Gen : public skeleton::unary_function< int, double > {
  double operator()( int index ) const {
    return static_cast< double >( ( rand( ) % 10 ) * ( index + 1 ) );
  }
} gen;

struct Div : public skeleton::binary_function< int, double, double > {
  double operator()( int index, double x ) const {
    return x / ( static_cast< double >( index ) + 1.0 );
  }
} divide;

struct Max : public skeleton::binary_function< double, double, double > {
   double operator()( double x, double y ) const {
     return ( x > y ) ? x : y;
   }
} max_func;
const double unit_of_max = -DBL_MAX;

struct Eqfilter : public skeleton::binary_function< double, double, int > {
  int operator()( double x, double y ) const {
    return ( x == y ) ? 1 : 0;
  }
} eqfilter;

int SketoMain( int argc, char **argv )
{
  if ( argc < 3 ) {
    skeleton::cout << "usage: " << argv[ 0 ] << " size loop" << std::endl;
    return 0;
  }

  const int size = std::atoi( argv[ 1 ] );
  const int loop = std::atoi( argv[ 2 ] );

  skeleton::time::initialize( ); // for time ekeeping

  for ( int i = 0; i < loop; i++ ) {
    // generate data
    srand( time( NULL ) );
    dist_list< double > *as = new dist_list< double >( gen, size );
    
    // calculate angle
    list_skeletons::mapindex_ow( divide, as );
    
    // calculate max angle before oneself
    dist_list< double > *bs = list_skeletons::prescan( max_func, unit_of_max, as );
    
    // if the angle equal max angle, you can see it.
    dist_list< int > *result = list_skeletons::zipwith( eqfilter, as, bs );
    delete as;
    delete bs;
    
    list_skeletons::print( result );
    delete result;
  }

  skeleton::time::set_time( "calculate" ); // for time ekeeping
  skeleton::time::print( ); // show the result of time ekeeping

  return 0;
}
