/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file variance.cpp
 * @brief This program compute average and variance.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#include <cstdlib>
#include <ctime>
#include "list_skeletons.h"

struct Gen : public skeleton::unary_function< int, double > {
  double operator()( int index ) const {
    return static_cast< double >( std::rand( ) % 100 );
  }
} gen;

struct Add : public skeleton::binary_function< double, double, double > {
  double operator()( double x, double y ) const { return x + y; }
} add;
const double add_unit = 0.0;

struct Sqr : public skeleton::unary_function< double, double > {
  double operator()( double x ) const { return x * x; }
} sqr;

struct Sub : public skeleton::unary_function< double, double > {
  double val;
  Sub( double val_ ) : val( val_ ) { }
  double operator()( double x ) const { return x - val; }
};

int SketoMain( int argc, char **argv )
{
  if ( argc < 3 ) {
    skeleton::cout << "usage: " << argv[ 0 ] << " size loop" << std::endl;
    return 0;
  }

  const int size = std::atoi( argv[ 1 ] );
  const int loop = std::atoi( argv[ 2 ] );

  std::srand( std::time ( NULL ) ); // initialize random seed

  skeleton::time::initialize( ); // for time ekeeping

  for ( int i = 0; i < loop; i++ ) {
    // generate data
    dist_list< double > *as = new dist_list< double >( gen, size );
    
    // calculate average
    double sum1 = list_skeletons::reduce( add, add_unit, as );
    double ave = sum1 / size;
    
    // calculate variance
    Sub sub( ave );
    list_skeletons::map_ow( sub, as );
    list_skeletons::map_ow( sqr, as );
    double sum2 = list_skeletons::reduce( add, add_unit, as );
    double var = sum2 / size;
    delete as;

    // output results
    skeleton::cout << "average: "  << ave << std::endl;
    skeleton::cout << "variance: " << var << std::endl;
  }

  skeleton::time::set_time( "calculate" ); // for time ekeeping
  skeleton::time::print( ); // show the result of time ekeeping

  return 0;
}
