/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/***
 * @file fnorm.cpp
 * @brief The Frobenius Norm.  
 * @author Kento EMOTO
 */
#include <cstdlib>
#include <iostream>
#include <cmath>
#include "matrix_skeletons.h"
#include "primitive_functions.h"
using namespace primitive_functions;

template<typename A>
class AddN : public skeleton::binary_function< A, A, A > {
public:
	void operator()(const A &i, const A &j, A *r) const {
		*r = i + j;
	}
	template<typename B>
	void operator()(const A &i, const A &j, A *r, B&) const {
		*r = i + j;
	}
};

double fnorm(const matrix<double> &mat){
	skeleton::time::initialize( );
	dist_matrix<double> dmat(&mat);
	skeleton::time::set_time( "after distribute" );
	dist_matrix<double> *dmat2 = matrix_skeletons::map(Square<double>(), &dmat);
	double *ss = matrix_skeletons::reduce(Add<double>(), Add<double>(), dmat2);
	double ret = sqrt(*ss);
	skeleton::time::set_time( "after computation" );
	skeleton::time::print( );
	delete ss;
	delete dmat2;
	return ret;
}

double fnormOpt(const matrix<double> &mat){
	skeleton::time::initialize( );
	dist_matrix<double> dmat(&mat);
	skeleton::time::set_time( "after distribute" );
	double *ss = matrix_skeletons::cata(Add<double>(), Add<double>(), Square<double>(), &dmat);
	double ret = sqrt(*ss);
	skeleton::time::set_time( "after computation (opt)" );
	skeleton::time::print( );
	delete ss;
	return ret;
}

double fnormN(const matrix<double> &mat){
	skeleton::time::initialize( );
	dist_matrix<double> dmat(&mat);
	skeleton::time::set_time( "after distribute" );
	dist_matrix<double> *dmat2 = matrix_skeletons::map(Square<double>(), &dmat);
	double *ss = matrix_skeletons::reduceDCN2(AddN<double>(), AddN<double>(), dmat2);
	double ret = sqrt(*ss);
	skeleton::time::set_time( "after computation" );
	skeleton::time::print( );
	delete ss;
	delete dmat2;
	return ret;
}

double fnormNOpt(const matrix<double> &mat){
	skeleton::time::initialize( );
	dist_matrix<double> dmat(&mat);
	skeleton::time::set_time( "after distribute" );
	double *ss = matrix_skeletons::cataDCN2(AddN<double>(), AddN<double>(), Square<double>(), &dmat);
	double ret = sqrt(*ss);
	skeleton::time::set_time( "after computation (opt)" );
	skeleton::time::print( );
	delete ss;
	return ret;
}


int SketoMain( int argc, char **argv )
{
	int n = 4000;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	int t = 0;
	if(argc > 2) t = atoi(argv[2]);
	matrix<double> mat(n, n);
	mat.generate(GenSub<double>());
	double norm;
	switch(t){
	case 0: norm = fnorm(mat); break;
	case 1: norm = fnormOpt(mat); break;
	case 2: norm = fnormN(mat); break;
	case 3: norm = fnormNOpt(mat); break;
	}
	skeleton::cout << "#procs = "<< skeleton::procs << std::endl;
	skeleton::cout << "matrix size = " << n << "x" << n << std::endl;
	skeleton::cout << "The Frobenius Norm = " << norm << std::endl;

	if(argc > 3){
		skeleton::cout << "the matrix = " << std::endl;
		skeleton::cout << mat<<std::endl;
	}
	return 0;
}
