/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/***
 * @file mulssp.cpp
 * @brief Maximum Upper-Left Submatrix Sum Problem.  by Kento EMOTO
 * @author Kento EMOTO
 */
#include <iostream>
#include <cstdlib>
#include "matrix_skeletons.h"
#include "primitive_functions.h"
using namespace primitive_functions;

double mulssp(const matrix<double> &mat){
	skeleton::time::initialize( );
	dist_matrix<double> dmat(&mat);
	skeleton::time::set_time( "after distribute" );
	dist_matrix<double> *dmat2 = matrix_skeletons::scan(Add<double>(), Add<double>(), &dmat);
	double *ss = matrix_skeletons::reduce(Max<double>(), Max<double>(), dmat2);
	double ret = *ss;
	skeleton::time::set_time( "after computation" );
	skeleton::time::print( );
	delete ss;
	delete dmat2;
	return ret;
}

double mulssp2(const matrix<double> &mat){
	skeleton::time::initialize( );
	dist_matrix<double> dmat(&mat);
	skeleton::time::set_time( "after distribute" );
	dist_matrix<double> *dmat2 = matrix_skeletons::scan2(Add<double>(), Add<double>(), &dmat);
	double *ss = matrix_skeletons::reduce(Max<double>(), Max<double>(), dmat2);
	double ret = *ss;
	skeleton::time::set_time( "after computation(scan2)" );
	skeleton::time::print( );
	delete ss;
	delete dmat2;
	return ret;
}

double mulssp3(const matrix<double> &mat){
	skeleton::time::initialize( );
	dist_matrix<double> dmat(&mat);
	skeleton::time::set_time( "after distribute" );
	double *ss = matrix_skeletons::acc(Max<double>(), Max<double>(), Snd<double, double>(), Add<double>(), Add<double>(), Id<double>(), &dmat);
	double ret = *ss;
	skeleton::time::set_time( "after computation(acc)" );
	skeleton::time::print( );
	delete ss;
	return ret;
}

int SketoMain( int argc, char **argv )
{

	int n = 1000;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	matrix<double> mat(n, n);
	mat.generate(GenSub<double>());
	double mulss;
	int k = 0;
	if(argc > 2) k = atoi(argv[2]);
	switch(k){
	case 1: mulss = mulssp2(mat); break;
	case 2: mulss = mulssp3(mat); break;
	default:
	case 0:	mulss = mulssp(mat); break;
	}
	skeleton::cout << "#procs = "<< skeleton::procs << std::endl;
	skeleton::cout << "matrix size = " << n << "x" << n << std::endl;
	skeleton::cout << "Maximum Upper-Left Submatrix Sum = " << mulss << std::endl;
	if(argc > 3){
		skeleton::cout << "the matrix = " << std::endl;
		skeleton::cout<< mat<<std::endl;
	}
	return 0;
}

 
