/**
 * @file: make_bin_tree.cpp
 * @brief Make a binary-tree representation of perfect binary tree
 *        for a given depth.
 */

#include <iostream>
#include <cstdlib>
#include <cstdio>

using namespace std;

#include "tree_util.h"

int pow2( int n )
{
  if ( n == 0 ) return 1;
  return 2 * pow2( n - 1 );
}

void fill( int* flags, int* values, int* index, int height )
{
  if ( height > 0 ) {
    flags[ *index ] = 0;
    values[ *index ] = rand( ) % 10;
    (*index)++;
    fill( flags, values, index, height - 1 );
    
    flags[ *index ] = 0;
    values[ *index ] = rand( ) % 10;
    (*index)++;
    fill( flags, values, index, height - 1 );
  }
  flags[ *index ] = 1;
  values[ *index ] = 0;
  (*index)++;
}

int main( int argc, char** argv )
{
  if ( argc < 1 + 4 ) {
    cerr << "usage: make_bin_tree <height> <seed> <prename> <nprocs>" << endl;
    return 1;
  }
  
  const int height = atoi( argv[ 1 ] );
  const int seed   = atoi( argv[ 2 ] );
  const string prename( argv[ 3 ] );

  srand( seed );
  int* flags = new int[ 2 * pow2( height ) - 1 ];
  int* values = new int[ 2 * pow2( height ) - 1 ];

  int index = 0;
  flags[ index ] = 0;
  values[ index ] = rand( ) % 10;
  index++;

  fill( flags, values, &index, height - 1 );

  flags[ index ] = 1;
  values[ index ] = 0;
  index++;

  node< int >* tree = tree_util::construct_tree_from_array( flags, values );
  for ( int i = 4; i < argc; i++ ) {
    int nprocs = atoi( argv[ i ] );
    string nprocs_str( argv[ i ] );
    string filename = prename + nprocs_str;
    tree_util::dist_and_write_to_file( tree, nprocs, filename.c_str( ) );
  }
}
