/**
 * @file: make_monadic_tree.cpp
 * @brief Make a binary-tree representation of perfect monadic tree
 *        for a given size
 */

#include <iostream>
#include <cstdlib>
#include <cstdio>

using namespace std;

#include "tree_util.h"

int main( int argc, char** argv )
{
  if ( argc < 1 + 4 ) {
    cerr << "usage: make_monadic_tree <size> <seed> <prename> <nprocs>" << endl;
    return 1;
  }
  
  const int size = atoi( argv[ 1 ] );
  const int seed = atoi( argv[ 2 ] );
  const string prename( argv[ 3 ] );

  srand( seed );
  int* flags = new int[ 2 * size + 1 ];
  int* values = new int[ 2 * size + 1 ];

  for ( int i = 0; i < size; i++ ) {
    flags[ i ] = 0;
    values[ i ] = rand( ) % 10;
  }
  for ( int i = size; i < size * 2 + 1; i++ ) {
    flags[ i ] = 1;
    values[ i ] = 0;
  }

  node< int >* tree = tree_util::construct_tree_from_array( flags, values );
  for ( int i = 4; i < argc; i++ ) {
    int nprocs = atoi( argv[ i ] );
    string nprocs_str( argv[ i ] );
    string filename = prename + nprocs_str;
    tree_util::dist_and_write_to_file( tree, nprocs, filename.c_str( ) );
  }
}
