/**
 * @file: make_bin_tree.cpp
 * @brief Make a binary-tree representation of perfect binary tree
 *        for a given depth.
 */

#include <iostream>
#include <cstdlib>
#include <cstdio>

using namespace std;

#include "tree_util.h"

int pow2( int n )
{
  if ( n == 0 ) return 1;
  return 2 * pow2( n - 1 );
}

void fill( int* flags, int* values, int* index, int height )
{
  if ( height > 0 ) {
    flags[ *index ] = 0;
    values[ *index ] = rand( ) % 10;
    (*index)++;
    fill( flags, values, index, height - 1 );
    
    flags[ *index ] = 0;
    values[ *index ] = rand( ) % 10;
    (*index)++;
    fill( flags, values, index, height - 1 );
  }
  flags[ *index ] = 1;
  values[ *index ] = 0;
  (*index)++;
}

void fill_null( node< int >* root )
{
  if ( root->left == NULL ) {
    root->left = new node< int >( 0 );
  } else {
    fill_null( root->left );
  }

  if ( root->right == NULL ) {
    root->right = new node< int >( 0 );
  } else {
    fill_null( root->right );
  }
}
    
int main( int argc, char** argv )
{
  if ( argc < 1 + 5 ) {
    cerr << "usage: make_bin_tree <size> <height> <seed> <prename> <nprocs>" << endl;
    return 1;
  }
  
  const int size   = atoi( argv[ 1 ] );
  const int height = atoi( argv[ 2 ] );
  const int seed   = atoi( argv[ 3 ] );
  const string prename( argv[ 4 ] );
  
  srand( seed );

  double tochild = 0.4;
  
  node< int >* tree = new node< int >( rand( ) % 10 );
  tree->left = new node< int >( rand( ) % 10 );
  for ( int i = 2; i < size; i++ ) {
    int h = 2; node< int >* root = tree->left;
    while( true ) {
      if ( h == height ) {
	if ( root->right == NULL ) {
	  root->right = new node< int >( rand( ) % 10 );
	  break;
	} else {
	  root = root->right;
	}
      } else {
	double d = rand( ) / (RAND_MAX + 1.0 );
	if ( d < tochild ) {
	  if ( root->left == NULL ) {
	    root->left = new node< int >( rand( ) % 10 );
	    break;
	  } else {
	    root = root->left;
	    h++;
	  }
	} else {
	  if ( root->right == NULL ) {
	    root->right = new node< int >( rand( ) % 10 );
	    break;
	  } else {
	    root = root->right;
	  }
	}
      }
    }
  }

  fill_null( tree );
  
  for ( int i = 5; i < argc; i++ ) {
    int nprocs = atoi( argv[ i ] );
    string nprocs_str( argv[ i ] );
    string filename = prename + nprocs_str;
    tree_util::dist_and_write_to_file( tree, nprocs, filename.c_str( ) );
  }
}
