#include <iostream>
#include <string>
#include <cassert>

using namespace std;

#include <rose_tree_skeletons.h>

struct func_1 : skeleton::unary_function< int, int > {
  int operator()( const int & ) const {
    return 1;
  }
};

struct func_add : skeleton::binary_function< int, int, int > {
  int operator( )( const int &x, const int &y ) const {
    return x + y;
  }
};

struct func_add1 : skeleton::unary_function< int, int > {
  int operator()( const int &x ) const {
    return x + 1;
  }
};

int mainRoutine( string inputfile ) {
  dist_rose_tree< int > *tree1
    = dist_rose_tree< int >::read_from_file( inputfile.c_str( ) );
  assert( tree1 );

  MPI_Barrier( MPI_COMM_WORLD );
  double start_time = MPI_Wtime( );
  
  dist_rose_tree< int > *tree2
    = rose_tree_skeletons::map( func_1( ), tree1 );
  if ( tree1 ) delete tree1;

  dist_rose_tree< int > *tree3
    = rose_tree_skeletons::uAcc_assoc( func_add( ), 0, tree2 );
  if ( tree2 ) delete tree2;

  dist_rose_tree< int > *tree4
    = rose_tree_skeletons::rAcc( func_add( ), 0, tree3 );
  if ( tree3 ) delete tree3;

  dist_rose_tree< int > *tree5
    = rose_tree_skeletons::map( func_add1( ), tree4 );
  if ( tree4 ) delete tree4;

  tree5->setroot( 0 );

  dist_rose_tree< int > *tree6
    = rose_tree_skeletons::dAcc( func_add( ), 0, tree5 );

  dist_rose_tree< int > *tree7
    = rose_tree_skeletons::zipwith( func_add( ), tree6, tree5 );
  if ( tree5 ) delete tree5;
  if ( tree6 ) delete tree6;

  //  assert( tree7->write_to_file( "output.txt" ) );
  MPI_Barrier( MPI_COMM_WORLD );
  double end_time = MPI_Wtime( );
  skeleton::cout << ( end_time - start_time ) << endl;
  
  if ( tree7 ) delete tree7;
  
  return 0;
}

int SketoMain( int argc, char** argv ) {
  string inputfile;
  if ( argc >= 1 + 1 ) {
    inputfile = string( argv[ 1 ] );
  } else {
    skeleton::cout << "usage: pon <inputfile>" << endl;
    return 1;
  }

  return mainRoutine( inputfile );
}

